# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import unittest

from libcloud.common.types import LibcloudError
from libcloud.test.secrets import STORAGE_S3_PARAMS
from libcloud.storage.drivers.minio import MinIOStorageDriver, MinIOConnectionAWS4


class MinIOStorageDriverTestCase(unittest.TestCase):
    driver_type = MinIOStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = "localhost"

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args, host=self.default_host)

    def setUp(self):
        self.driver = self.create_driver()

    def test_connection_class_type(self):
        self.assertEqual(self.driver.connectionCls, MinIOConnectionAWS4)

    def test_connection_class_default_host(self):
        self.assertEqual(self.driver.connectionCls.host, self.default_host)
        self.assertEqual(self.driver.connectionCls.port, 443)

    def test_connection_class_host_port_secure(self):
        host = "127.0.0.3"
        port = 9000
        driver = self.driver_type(*self.driver_args, host=host, port=port, secure=False)
        self.assertEqual(driver.connectionCls.host, host)
        self.assertEqual(driver.connection.port, 9000)
        self.assertEqual(driver.connection.secure, False)

        host = "127.0.0.4"
        port = 9000
        driver = self.driver_type(*self.driver_args, host=host, port=port, secure=True)
        self.assertEqual(driver.connectionCls.host, host)
        self.assertEqual(driver.connection.port, 9000)
        self.assertEqual(driver.connection.secure, True)

    def test_empty_host_error(self):
        self.assertRaisesRegex(
            LibcloudError,
            "host argument is required",
            self.driver_type,
            *self.driver_args,
        )


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest

from libcloud.common.base import Connection, ConnectionKey, ConnectionUserAndKey
from libcloud.common.types import LibcloudError
from libcloud.compute.base import (
    Node,
    NodeSize,
    NodeImage,
    NodeDriver,
    StorageVolume,
    NodeAuthSSHKey,
    NodeAuthPassword,
)
from libcloud.compute.types import StorageVolumeState


class FakeDriver:
    type = 0


class BaseTests(unittest.TestCase):
    def test_base_node(self):
        Node(id=0, name=0, state=0, public_ips=0, private_ips=0, driver=FakeDriver())

    def test_base_node_size(self):
        NodeSize(id=0, name=0, ram=0, disk=0, bandwidth=0, price=0, driver=FakeDriver())

    def test_base_node_image(self):
        NodeImage(id=0, name=0, driver=FakeDriver())

    def test_base_storage_volume(self):
        StorageVolume(
            id="0",
            name="0",
            size=10,
            driver=FakeDriver(),
            state=StorageVolumeState.AVAILABLE,
        )

    def test_base_node_driver(self):
        NodeDriver("foo")

    def test_base_connection_key(self):
        ConnectionKey("foo")

    def test_base_connection_userkey(self):
        ConnectionUserAndKey("foo", "bar")

    def test_base_connection_timeout(self):
        Connection(timeout=10)


class TestValidateAuth(unittest.TestCase):
    def test_get_auth_ssh(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["ssh_key"]}
        auth = NodeAuthSSHKey("pubkey...")
        self.assertEqual(auth, n._get_and_check_auth(auth))

    def test_get_auth_ssh_but_given_password(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["ssh_key"]}
        auth = NodeAuthPassword("password")
        self.assertRaises(LibcloudError, n._get_and_check_auth, auth)

    def test_get_auth_password(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["password"]}
        auth = NodeAuthPassword("password")
        self.assertEqual(auth, n._get_and_check_auth(auth))

    def test_get_auth_password_but_given_ssh_key(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["password"]}
        auth = NodeAuthSSHKey("publickey")
        self.assertRaises(LibcloudError, n._get_and_check_auth, auth)

    def test_get_auth_default_ssh_key(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["ssh_key"]}
        self.assertEqual(None, n._get_and_check_auth(None))

    def test_get_auth_default_password(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["password"]}
        auth = n._get_and_check_auth(None)
        self.assertTrue(isinstance(auth, NodeAuthPassword))

    def test_get_auth_default_no_feature(self):
        n = NodeDriver("foo")
        self.assertEqual(None, n._get_and_check_auth(None))

    def test_get_auth_generates_password_but_given_nonsense(self):
        n = NodeDriver("foo")
        n.features = {"create_node": ["generates_password"]}
        auth = "nonsense"
        self.assertRaises(LibcloudError, n._get_and_check_auth, auth)

    def test_get_auth_no_features_but_given_nonsense(self):
        n = NodeDriver("foo")
        auth = "nonsense"
        self.assertRaises(LibcloudError, n._get_and_check_auth, auth)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Subclass for httplib.HTTPSConnection with optional certificate name
verification, depending on libcloud.security settings.
"""

import os
import warnings

import requests
from requests.adapters import HTTPAdapter

import libcloud.security
from libcloud.utils.py3 import urlparse

try:
    # requests no longer vendors urllib3 in newer versions
    # https://github.com/python/typeshed/issues/6893#issuecomment-1012511758
    from urllib3.poolmanager import PoolManager
except ImportError:
    from requests.packages.urllib3.poolmanager import PoolManager  # type: ignore


__all__ = ["LibcloudBaseConnection", "LibcloudConnection"]

ALLOW_REDIRECTS = 1

# Default timeout for HTTP requests in seconds
DEFAULT_REQUEST_TIMEOUT = 60

HTTP_PROXY_ENV_VARIABLE_NAME = "http_proxy"
HTTPS_PROXY_ENV_VARIABLE_NAME = "https_proxy"


class SignedHTTPSAdapter(HTTPAdapter):
    def __init__(self, cert_file, key_file):
        self.cert_file = cert_file
        self.key_file = key_file
        super().__init__()

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            cert_file=self.cert_file,
            key_file=self.key_file,
        )


class LibcloudBaseConnection:
    """
    Base connection class to inherit from.

    Note: This class should not be instantiated directly.
    """

    session = None

    proxy_scheme = None
    proxy_host = None
    proxy_port = None

    proxy_username = None
    proxy_password = None

    http_proxy_used = False

    ca_cert = None

    def __init__(self):
        self.session = requests.Session()

    def set_http_proxy(self, proxy_url):
        """
        Set a HTTP proxy which will be used with this connection.

        :param proxy_url: Proxy URL (e.g. http://<hostname>:<port> without
                          authentication and
                          http://<username>:<password>@<hostname>:<port> for
                          basic auth authentication information.
        :type proxy_url: ``str``
        """
        result = self._parse_proxy_url(proxy_url=proxy_url)

        scheme = result[0]
        host = result[1]
        port = result[2]
        username = result[3]
        password = result[4]

        self.proxy_scheme = scheme
        self.proxy_host = host
        self.proxy_port = port
        self.proxy_username = username
        self.proxy_password = password
        self.http_proxy_used = True

        self.session.proxies = {
            "http": proxy_url,
            "https": proxy_url,
        }

    def _parse_proxy_url(self, proxy_url):
        """
        Parse and validate a proxy URL.

        :param proxy_url: Proxy URL (e.g. http://hostname:3128)
        :type proxy_url: ``str``

        :rtype: ``tuple`` (``scheme``, ``hostname``, ``port``)
        """
        parsed = urlparse.urlparse(proxy_url)

        if parsed.scheme not in ("http", "https"):
            raise ValueError("Only http and https proxies are supported")

        if not parsed.hostname or not parsed.port:
            raise ValueError(
                "proxy_url must be in the following format: " "<scheme>://<proxy host>:<proxy port>"
            )

        proxy_scheme = parsed.scheme
        proxy_host, proxy_port = parsed.hostname, parsed.port

        netloc = parsed.netloc

        if "@" in netloc:
            username_password = netloc.split("@", 1)[0]
            split = username_password.split(":", 1)

            if len(split) < 2:
                raise ValueError("URL is in an invalid format")

            proxy_username, proxy_password = split[0], split[1]
        else:
            proxy_username = None
            proxy_password = None

        return (proxy_scheme, proxy_host, proxy_port, proxy_username, proxy_password)

    def _setup_verify(self):
        self.verify = libcloud.security.VERIFY_SSL_CERT

    def _setup_ca_cert(self, **kwargs):
        # simulating keyword-only argument in Python 2
        ca_certs_path = kwargs.get("ca_cert", libcloud.security.CA_CERTS_PATH)

        if self.verify is False:
            pass
        else:
            if isinstance(ca_certs_path, list):
                msg = (
                    "Providing a list of CA trusts is no longer supported "
                    "since libcloud 2.0. Using the first element in the list. "
                    "See http://libcloud.readthedocs.io/en/latest/other/"
                    "changes_in_2_0.html#providing-a-list-of-ca-trusts-is-no-"
                    "longer-supported"
                )
                warnings.warn(msg, DeprecationWarning)
                self.ca_cert = ca_certs_path[0]
            else:
                self.ca_cert = ca_certs_path

    def _setup_signing(self, cert_file=None, key_file=None):
        """
        Setup request signing by mounting a signing
        adapter to the session
        """
        self.session.mount("https://", SignedHTTPSAdapter(cert_file, key_file))


class LibcloudConnection(LibcloudBaseConnection):
    timeout = None
    host = None
    response = None

    def __init__(self, host, port, secure=None, **kwargs):
        scheme = "https" if secure is not None and secure else "http"
        self.host = "{}://{}{}".format(
            "https" if port == 443 else scheme,
            host,
            ":{}".format(port) if port not in (80, 443) else "",
        )

        # Support for HTTP(s) proxy
        # NOTE: We always only use a single proxy (either HTTP or HTTPS)
        https_proxy_url_env = os.environ.get(HTTPS_PROXY_ENV_VARIABLE_NAME, None)
        http_proxy_url_env = os.environ.get(HTTP_PROXY_ENV_VARIABLE_NAME, https_proxy_url_env)

        # Connection argument has precedence over environment variables
        proxy_url = kwargs.pop("proxy_url", http_proxy_url_env)

        self._setup_verify()
        self._setup_ca_cert()

        LibcloudBaseConnection.__init__(self)

        self.session.timeout = kwargs.pop("timeout", DEFAULT_REQUEST_TIMEOUT)

        if "cert_file" in kwargs or "key_file" in kwargs:
            self._setup_signing(**kwargs)

        if proxy_url:
            self.set_http_proxy(proxy_url=proxy_url)

    @property
    def verification(self):
        """
        The option for SSL verification given to underlying requests
        """
        return self.ca_cert if self.ca_cert is not None else self.verify

    def request(self, method, url, body=None, headers=None, raw=False, stream=False, hooks=None):
        url = urlparse.urljoin(self.host, url)
        headers = self._normalize_headers(headers=headers)

        self.response = self.session.request(
            method=method.lower(),
            url=url,
            data=body,
            headers=headers,
            allow_redirects=ALLOW_REDIRECTS,
            stream=stream,
            verify=self.verification,
            timeout=self.session.timeout,
            hooks=hooks,
        )

    def prepared_request(self, method, url, body=None, headers=None, raw=False, stream=False):
        headers = self._normalize_headers(headers=headers)

        req = requests.Request(method, "".join([self.host, url]), data=body, headers=headers)

        prepped = self.session.prepare_request(req)

        self.response = self.session.send(
            prepped,
            stream=stream,
            verify=self.ca_cert if self.ca_cert is not None else self.verify,
        )

    def getresponse(self):
        return self.response

    def getheaders(self):
        # urlib decoded response body, libcloud has a bug
        # and will not check if content is gzipped, so let's
        # remove headers indicating compressed content.
        if "content-encoding" in self.response.headers:
            del self.response.headers["content-encoding"]
        return self.response.headers

    @property
    def status(self):
        return self.response.status_code

    @property
    def reason(self):
        return None if self.response.status_code > 400 else self.response.text

    def connect(self):  # pragma: no cover
        pass

    def read(self):
        return self.response.content

    def close(self):  # pragma: no cover
        # return connection back to pool
        self.response.close()

    def _normalize_headers(self, headers):
        headers = headers or {}

        # all headers should be strings
        for key, value in headers.items():
            if isinstance(value, (int, float)):
                headers[key] = str(value)

        return headers


class HttpLibResponseProxy:
    """
    Provides a proxy pattern around the :class:`requests.Reponse`
    object to a :class:`httplib.HTTPResponse` object
    """

    def __init__(self, response):
        self._response = response

    def read(self, amt=None):
        return self._response.text

    def getheader(self, name, default=None):
        """
        Get the contents of the header name, or default
        if there is no matching header.
        """
        if name in self._response.headers.keys():
            return self._response.headers[name]
        else:
            return default

    def getheaders(self):
        """
        Return a list of (header, value) tuples.
        """
        return list(self._response.headers.items())

    @property
    def status(self):
        return self._response.status_code

    @property
    def reason(self):
        return self._response.reason

    @property
    def version(self):
        # requests doesn't expose this
        return "11"

    @property
    def body(self):
        # NOTE: We use property to avoid saving whole response body into RAM
        # See https://github.com/apache/libcloud/pull/1132 for details
        return self._response.content



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl
import socket
from unittest.mock import Mock, MagicMock, patch

from libcloud.test import unittest
from libcloud.common.base import Connection
from libcloud.utils.retry import TRANSIENT_SSL_ERROR

CONFLICT_RESPONSE_STATUS = [
    ("status", "429"),
    ("reason", "CONFLICT"),
    ("retry_after", "3"),
    ("content-type", "application/json"),
]
SIMPLE_RESPONSE_STATUS = ("HTTP/1.1", 429, "CONFLICT")


@patch("os.environ", {"LIBCLOUD_RETRY_FAILED_HTTP_REQUESTS": True})
class FailedRequestRetryTestCase(unittest.TestCase):
    def test_retry_connection(self):
        con = Connection(timeout=0.2, retry_delay=0.1)
        con.connection = Mock()
        connect_method = "libcloud.common.base.Connection.request"

        with patch(connect_method) as mock_connect:
            try:
                mock_connect.side_effect = socket.gaierror("")
                con.request("/")
            except socket.gaierror:
                pass

    def test_retry_connection_ssl_error(self):
        conn = Connection(timeout=0.2, retry_delay=0.1)

        with patch.object(conn, "connect", Mock()):
            with patch.object(conn, "connection") as connection:
                connection.request = MagicMock(
                    __name__="request", side_effect=ssl.SSLError(TRANSIENT_SSL_ERROR)
                )

                self.assertRaises(ssl.SSLError, conn.request, "/")
                self.assertGreater(connection.request.call_count, 1)


if __name__ == "__main__":
    unittest.main()



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from libcloud.compute.base import (
    Node,
    KeyPair,
    NodeImage,
    NodeState,
    NodeDriver,
    NodeLocation,
    StorageVolume,
    VolumeSnapshot,
    StorageVolumeState,
)
from libcloud.utils.iso8601 import parse_date
from libcloud.common.gridscale import GridscaleBaseDriver, GridscaleConnection
from libcloud.compute.providers import Provider


class GridscaleIp:
    """
    Ip Object

    :param id: uuid
    :type id: ``str``
    :param family: family of ip (v4 or v6)
    :type family: ``str``
    :param prefix: prefix of ip
    :type prefix: ``str``
    :param ip_address: Ip address
    :type ip_address: ``str``
    :param create_time: Time ip was created
    :type create_time: ``str``
    """

    def __init__(self, id, family, prefix, create_time, address, extra=None):
        self.id = id
        self.family = family
        self.prefix = prefix
        self.create_time = create_time
        self.ip_address = address
        self.extra = extra or {}

    def __repr__(self):
        return "Ip: id={}, family={}, prefix={}, create_time={}, " "ip_address={}".format(
            self.id, self.family, self.prefix, self.create_time, self.ip_address
        )


class GridscaleNetwork:
    """
    Network Object

    :param id: uuid
    :type id: ``str``
    :param name: Name of Network
    :type name: ``str``
    :param status: Network status
    :type status: ``str``
    :param relations: object related to network
    :type relations: ``object``
    :param create_time: Time Network was created
    :type create_time: ``str``
    """

    def __init__(self, id, name, status, create_time, relations):
        self.id = id
        self.name = name
        self.status = status
        self.create_time = create_time
        self.relations = relations

    def __repr__(self):
        return "Network: id={}, name={}, status={}, create_time={}, " "relations={}".format(
            self.id, self.name, self.status, self.create_time, self.relations
        )


class GridscaleNodeDriver(GridscaleBaseDriver, NodeDriver):
    """
    create and entry in libcloud/compute/providers for gridscale
    """

    connectionCls = GridscaleConnection
    type = Provider.GRIDSCALE
    name = "Gridscale"
    api_name = "gridscale"
    website = "https://gridscale.io"
    features = {"create_node": ["ssh_key"]}

    def __init__(self, user_id, key, **kwargs):
        super().__init__(user_id, key, **kwargs)

    def list_nodes(self):
        """
        List all nodes.

        :return: List of node objects
        :rtype: ``list`` of :class:`.Node`
        """
        result = self._sync_request(data=None, endpoint="objects/servers/")
        nodes = []
        for key, value in self._get_response_dict(result).items():
            node = self._to_node(value)
            nodes.append(node)

        return sorted(nodes, key=lambda sort: sort.created_at)

    def list_locations(self):
        """
        List all available data centers.

        :return: List of node location objects
        :rtype: ``list`` of :class:`.NodeLocation`
        """
        locations = []
        result = self._sync_request(endpoint="objects/locations/")
        for key, value in self._get_response_dict(result).items():
            location = self._to_location(value)
            locations.append(location)

        return sorted(locations, key=lambda nod: nod.id)

    def list_volumes(self):
        """
        List all volumes.

        :return: List of StorageVolume object
        :rtype: ``list`` of :class:`.StorageVolume`
        """
        volumes = []
        result = self._sync_request(endpoint="objects/storages/")
        for key, value in self._get_response_dict(result).items():
            volume = self._to_volume(value)
            volumes.append(volume)
        return sorted(volumes, key=lambda sort: sort.extra["create_time"])

    def ex_list_networks(self):
        """
        List all networks.

        :return: List of objects.
        :rtype: ``list`` of :class:`.GridscaleNetwork`
        """
        networks = []
        result = self._sync_request(endpoint="objects/networks/")
        for key, value in self._get_response_dict(result).items():
            network = self._to_network(value)
            networks.append(network)
        return sorted(networks, key=lambda sort: sort.create_time)

    def list_volume_snapshots(self, volume):
        """
        Lists all snapshots for storage volume.

        :param volume: storage the snapshot is attached to
        :type volume: :class:`.StorageVolume`

        :return: Snapshots
        :rtype: ``list`` of :class:`.VolumeSnapshot`
        """
        snapshots = []
        result = self._sync_request(endpoint="objects/storages/" "{}/snapshots".format(volume.id))
        for key, value in self._get_response_dict(result).items():
            snapshot = self._to_volume_snapshot(value)
            snapshots.append(snapshot)
        return sorted(snapshots, key=lambda snapshot: snapshot.created)

    def ex_list_ips(self):
        """
        Lists all IPs available.

        :return: List of IP objects.
        :rtype: ``list`` of :class:`.GridscaleIp`
        """
        ips = []
        result = self._sync_request(endpoint="objects/ips/")
        for key, value in self._get_response_dict(result).items():
            ip = self._to_ip(value)
            ips.append(ip)
        return ips

    def list_images(self):
        """
        List images.

        :return: List of node image objects
        :rtype: ``list`` of :class:`.NodeImage`
        """
        templates = []
        result = self._sync_request(endpoint="objects/templates")
        for key, value in self._get_response_dict(result).items():
            template = self._to_node_image(value)
            templates.append(template)
        return sorted(templates, key=lambda sort: sort.name)

    def create_node(self, name, size, image, location, ex_ssh_key_ids=None, **kwargs):
        """
        Create a simple node  with a name, cores, memory at the designated
        location.

        :param name: Name of the server.
        :type name: ``str``

        :param size: Nodesize object.
        :type size: :class:`.NodeSize`

        :param image: OS image to attach to the storage.
        :type image: :class:`.GridscaleTemplate`

        :param location: The data center to create a node in.
        :type location: :class:`.NodeLocation`

        :keyword ex_ssh_key_ids: List of SSH key IDs to add to the server.
        :type ex_ssh_key_ids: ``list`` of ``str``

        :return: The newly created Node.
        :rtype: :class:`.Node`

        """

        if size.ram % 1024 != 0:
            raise Exception("Value not accepted. Use a multiple of 1024 e.g." "1024, 2048, 3072...")
        data = {
            "name": name,
            "cores": size.extra["cores"],
            "memory": int(size.ram / 1024),
            "location_uuid": location.id,
        }
        self.connection.async_request("objects/servers/", data=data, method="POST")

        node = self._to_node(
            self._get_resource(
                "servers", self.connection.poll_response_initial.object["object_uuid"]
            )
        )

        volume = self._create_volume_from_template(
            name=image.extra["ostype"],
            size=size.disk,
            location=location,
            template={"template_uuid": image.id, "sshkeys": ex_ssh_key_ids},
        )

        ip = self.ex_create_ip(4, location, name + "_ip")

        self.attach_volume(node, volume)
        self.ex_link_ip_to_node(node, ip)
        self.ex_link_network_to_node(node, self.ex_list_networks()[0])
        self.ex_start_node(node)

        return self._to_node(self._get_resource("servers", node.id))

    def ex_create_ip(self, family, location, name):
        """
        Create either an ip_v4 ip or a ip_v6.

        :param family: Defines if the ip is v4 or v6 with int 4 or int 6.
        :type family: ``int``

        :param location: Defines which datacenter the created ip
                         responds with.
        :type location: :class:`.NodeLocation`

        :param name: Name of your Ip.
        :type name: ``str``

        :return: Ip
        :rtype: :class:`.GridscaleIp`
        """
        self.connection.async_request(
            "objects/ips/",
            data={"name": name, "family": family, "location_uuid": location.id},
            method="POST",
        )

        return self._to_ip(
            self._get_resource("ips", self.connection.poll_response_initial.object["object_uuid"])
        )

    def ex_create_networks(self, name, location):
        """
        Create a network at the data center location.

        :param name: Name of the network.
        :type name: ``str``

        :param location: Location.
        :type location: :class:`.NodeLocation`

        :return: Network.
        :rtype: :class:`.GridscaleNetwork`
        """
        self.connection.async_request(
            "objects/networks",
            data={"name": name, "location_uuid": location.id},
            method="POST",
        )

        return self._to_network(
            self._get_resource(
                "network", self.connection.poll_response_initial.object["object_uuid"]
            )
        )

    def create_volume(self, size, name, location=None, snapshot=None):
        """
        Create a new volume.

        :param size: Integer in GB.
        :type size: ``int``

        :param name: Name of the volume.
        :type name: ``str``

        :param location: The server location.
        :type location: :class:`.NodeLocation`

        :param snapshot:  Snapshot from which to create the new
                          volume.  (optional)
        :type snapshot: :class:`.VolumeSnapshot`

        :return: Newly created StorageVolume.
        :rtype: :class:`.StorageVolume`
        """

        return self._create_volume_from_template(size, name, location)

    def _create_volume_from_template(self, size, name, location=None, template=None):
        """
        create Storage

        :param name: name of your Storage unit
        :type name: ``str``

        :param size: Integer in GB.
        :type size: ``int``

        :param location: your server location
        :type location: :class:`.NodeLocation`

        :param template: template to shape the storage capacity to
        :type template: ``dict``

        :return: newly created StorageVolume
        :rtype: :class:`.GridscaleVolumeStorage`
        """
        template = template
        self.connection.async_request(
            "objects/storages/",
            data={
                "name": name,
                "capacity": size,
                "location_uuid": location.id,
                "template": template,
            },
            method="POST",
        )

        return self._to_volume(
            self._get_resource(
                "storages", self.connection.poll_response_initial.object["object_uuid"]
            )
        )

    def create_volume_snapshot(self, volume, name):
        """
        Creates a snapshot of the current state of your volume,
        you can rollback to.

        :param volume: Volume you want to create a snapshot of.
        :type volume: :class:`.StorageVolume`

        :param name: Name of the snapshot.
        :type name: ``str``

        :return: VolumeSnapshot.
        :rtype: :class:`.VolumeSnapshot`
        """
        self.connection.async_request(
            "objects/storages/{}/snapshots".format(volume.id),
            data={"name": name},
            method="POST",
        )

        return self._to_volume_snapshot(
            self._get_resource(
                "storages/{}/snapshots".format(volume.id),
                self.connection.poll_response_initial.object["object_uuid"],
            )
        )

    def create_image(self, node, name):
        """
        Creates an image from a node object.

        :param node: Node to run the task on.
        :type node: :class:`.Node`

        :param name: Name for new image.
        :type name: ``str``

        :return: NodeImage.
        :rtype: :class:`.NodeImage`
        """
        storage_dict = node.extra["relations"]["storages"][0]
        snapshot_uuid = ""
        if storage_dict["bootdevice"] is True:
            self.connection.async_request(
                "objects/storages/{}/snapshots/".format(storage_dict["object_uuid"]),
                data={"name": name + "_snapshot"},
                method="POST",
            )

            snapshot_uuid = self.connection.poll_response_initial.object["object_uuid"]

            self.connection.async_request(
                "objects/templates/",
                data={"name": name, "snapshot_uuid": snapshot_uuid},
                method="POST",
            )

            snapshot_dict = self._get_response_dict(
                self._sync_request(
                    endpoint="objects/storages/{}/snapshots/{}".format(
                        storage_dict["object_uuid"], snapshot_uuid
                    )
                )
            )

            self.destroy_volume_snapshot(self._to_volume_snapshot(snapshot_dict))

        return self._to_node_image(
            self._get_resource(
                "templates", self.connection.poll_response_initial.object["object_uuid"]
            )
        )

    def destroy_node(self, node, ex_destroy_associated_resources=False):
        """
        Destroy node.

        :param node: Node object.
        :type node: :class:`.Node`

        :param ex_destroy_associated_resources: True to destroy associated
        resources such as storage volumes and IPs.
        :type ex_destroy_associated_resources: ``bool``

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        if ex_destroy_associated_resources:
            associated_volumes = self.ex_list_volumes_for_node(node=node)
            associated_ips = self.ex_list_ips_for_node(node=node)

        # 1. Delete the server itself
        result = self._sync_request(endpoint="objects/servers/{}".format(node.id), method="DELETE")

        # 2. Destroy associated resources (if requested)
        if ex_destroy_associated_resources:
            for volume in associated_volumes:
                self.destroy_volume(volume=volume)

            for ip in associated_ips:
                self.ex_destroy_ip(ip=ip)

        return result.status == 204

    def destroy_volume(self, volume):
        """
        Delete volume.

        :param volume: Volume to be destroyed.
        :type volume: :class:`.StorageVolume`

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        result = self._sync_request(
            endpoint="objects/storages/{}".format(volume.id), method="DELETE"
        )
        return result.status == 204

    def ex_destroy_ip(self, ip):
        """
        Delete an ip.

        :param ip: IP object.
        :type ip: :class:`.GridscaleIp`

        :return: ``True`` if delete_image was successful, ``False`` otherwise.
        :rtype: ``bool``
        """
        result = self._sync_request(endpoint="objects/ips/{}".format(ip.id), method="DELETE")
        return result.status == 204

    def destroy_volume_snapshot(self, snapshot):
        """
        Destroy a snapshot.

        :param snapshot: The snapshot to delete.
        :type snapshot: :class:'.VolumeSnapshot`

        :return: True if the destroy was successful, otherwise False.
        :rtype: ``bool``
        """
        result = self._sync_request(
            endpoint="objects/storages/"
            "{}/snapshots/{}/".format(snapshot.extra["parent_uuid"], snapshot.id),
            method="DELETE",
        )
        return result.status == 204

    def ex_destroy_network(self, network):
        """
        Delete network.

        :param network: Network object.
        :type network: :class:`.GridscaleNetwork`

        :return: ``True`` if destroyed successfully, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            endpoint="objects/networks/{}".format(network.id), method="DELETE"
        )
        return result.status == 204

    def delete_image(self, node_image):
        """
        Destroy an image.

        :param node_image: Node image object.
        :type node_image: :class:`.NodeImage`

        :return: True if the destroy was successful, otherwise False
        :rtype: ``bool``

        """
        result = self._sync_request(
            endpoint="objects/templates/{}".format(node_image.id), method="DELETE"
        )
        return result.status == 204

    def ex_rename_node(self, node, name):
        """
        Modify node name.

        :param name: New node name.
        :type name: ``str``

        :param node: Node
        :type node: :class:`.Node`

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            data={"name": name},
            endpoint="objects/servers/{}".format(node.id),
            method="PATCH",
        )
        return result.status == 204

    def ex_rename_volume(self, volume, name):
        """
        Modify storage volume name

        :param volume: Storage.
        :type volume: :class:.`StorageVolume`

        :param name: New storage name.
        :type name: ``str``

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            data={"name": name},
            endpoint="objects/storages/{}".format(volume.id),
            method="PATCH",
        )
        return result.status == 204

    def ex_rename_network(self, network, name):
        """
        Modify networks name.

        :param network: Network.
        :type network: :class:`.GridscaleNetwork`

        :param name: New network name.
        :type name: ``str``

        :return: ``True`` or ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            data={"name": name},
            endpoint="objects/networks/{}".format(network.id),
            method="PATCH",
        )
        return result.status == 204

    def reboot_node(self, node, ex_sleep_interval=3):
        """
        Reboot a node.

        :param node: Node object.
        :type node: :class:`.Node`

        :return: True if the reboot was successful, otherwise False.
        :rtype: ``bool``

        :keyword ex_sleep_interval: time to let the shutdown process finish
        :type ex_sleep_interval: ``int``

        """

        if node.extra["power"] is True:
            data = dict({"power": False})
            self._sync_request(
                data=data,
                endpoint="objects/servers/{}/power".format(node.id),
                method="PATCH",
            )
            time.sleep(ex_sleep_interval)

            data = dict({"power": True})
            self._sync_request(
                data=data,
                endpoint="objects/servers/{}/power".format(node.id),
                method="PATCH",
            )
            return True

        else:
            return False

    def import_key_pair_from_string(self, name, key_material):
        data = {"name": name, "sshkey": key_material}
        result = self._sync_request(endpoint="objects/sshkeys/", method="POST", data=data)
        key = self._to_key(result.object, name=name, sshkey=key_material)
        return key

    def list_key_pairs(self):
        """
        List all the available key pair objects.

        :rtype: ``list``of :class:`.KeyPair` objects
        """
        keys = []
        result = self._sync_request(endpoint="objects/sshkeys/")
        for key, value in self._get_response_dict(result).items():
            key = self._to_key(value)
            keys.append(key)
        return keys

    def get_image(self, image_id):
        """
        Get an image based on an image_id.

        :param image_id: Image identifier.
        :type image_id: ``str``

        :return: A NodeImage object.
        :rtype: :class:`.NodeImage`
        """

        response_dict = self._get_response_dict(
            self._sync_request(endpoint="/objects/templates/{}".format(image_id))
        )

        return self._to_node_image(response_dict)

    def start_node(self, node):
        result = self._sync_request(
            data={"power": True},
            endpoint="objects/servers/{}/power".format(node.id),
            method="PATCH",
        )

        return result.status == 204

    def ex_start_node(self, node):
        # NOTE: This method is here for backward compatibility reasons after
        # this method was promoted to be part of the standard compute API in
        # Libcloud v2.7.0
        return self.start_node(node=node)

    def ex_link_isoimage_to_node(self, node, isoimage):
        """
        link and isoimage to a node

        :param node: Node you want to link the iso image to
        :type node: ``object``

        :param isoimage: isomiage you want to link
        :type isoimage: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(
            data={"object_uuid": isoimage.id},
            endpoint="objects/servers/{}/isoimages/".format(node.id),
            method="POST",
        )
        return result

    def attach_volume(self, node, volume):
        """
         Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`.Node`

        :param volume: Volume to attach.
        :type volume: :class:`.StorageVolume`

        :rytpe: ``bool``
        """
        result = self._sync_request(
            data={"object_uuid": volume.id},
            endpoint="objects/servers/{}/storages/".format(node.id),
            method="POST",
        )
        return result.status == 204

    def ex_link_network_to_node(self, node, network):
        """
        Link a network to a node.

        :param node: Node object to link networks to.
        :type node: :class:`.Node`

        :param network: Network you want to link.
        :type network: :class:`.GridscaleNetwork`

        :return: ``True`` if linked successfully, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            data={"object_uuid": network.id},
            endpoint="objects/servers/{}/networks/".format(node.id),
            method="POST",
        )
        return result.status == 204

    def ex_link_ip_to_node(self, node, ip):
        """
        links a existing ip with a node

        :param node: node object
        :type node: ``object``

        :param ip: ip object
        :type ip: ``object``

        :return: Request ID
        :rtype: ``str``
        """
        result = self._sync_request(
            data={"object_uuid": ip.id},
            endpoint="objects/servers/{}/ips/".format(node.id),
            method="POST",
        )
        return result

    def ex_unlink_isoimage_from_node(self, node, isoimage):
        """
        unlink isoimages from server

        :param node: node you want to unlink the image from
        :type node: ``object``

        :param isoimage: isoimage you want to unlink
        :type isoimage: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(
            endpoint="objects/servers/{}/isoimages/{}".format(node.id, isoimage.id),
            method="DELETE",
        )
        return result

    def ex_unlink_ip_from_node(self, node, ip):
        """
        unlink ips from server

        :param node: node you want to unlink the ip from
        :type node: ``object``

        :param ip: the ip you want to unlink
        :type ip: ``object``

        :return: None -> success
        :rtype: ``None``
        """
        result = self._sync_request(
            endpoint="objects/servers/{}/ips/{}".format(node.id, ip.id), method="DELETE"
        )
        return result

    def ex_unlink_network_from_node(self, node, network):
        """
        Unlink network from node.

        :param node: Node you want to unlink from network.
        :type node: :class:`.Node`

        :param network: Network you want to unlink.
        :type network: :class:`.GridscaleNetwork

        :return: ``True`` if unlink was successful, otherwise ``False``
        :rtype: ``bool``
        """
        result = self._sync_request(
            endpoint="objects/servers/{}/networks/{}".format(node.id, network.id),
            method="DELETE",
        )
        return result.status == 204

    def detach_volume(self, volume):
        """
        Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`.StorageVolume`

        :rtype: ``bool``
        """
        node = volume.extra["relations"]["servers"][0]
        result = self._sync_request(
            endpoint="objects/servers/{}/storages/{}".format(node["object_uuid"], volume.id),
            method="DELETE",
        )
        return result.status == 204

    def ex_storage_rollback(self, volume, snapshot, rollback):
        """
        initiate a rollback on your storage

        :param volume: storage uuid
        :type volume: ``string``

        :param snapshot: snapshot uuid
        :type snapshot: ``string``

        :param rollback: variable
        :type rollback: ``bool``

        :return: RequestID
        :rtype: ``str``
        """
        result = self._sync_request(
            data={"rollback": rollback},
            endpoint="objects/storages/{}/snapshots/" "{}/rollback".format(volume.id, snapshot.id),
            method="PATCH",
        )
        return result

    def ex_list_volumes_for_node(self, node):
        """
        Return a list of associated volumes for the provided node.

        :rtype: ``list`` of :class:`StorageVolume`
        """
        volumes = self.list_volumes()

        result = []
        for volume in volumes:
            related_servers = volume.extra.get("relations", {}).get("servers", [])
            for server in related_servers:
                if server["object_uuid"] == node.id:
                    result.append(volume)

        return result

    def ex_list_ips_for_node(self, node):
        """
        Return a list of associated IPs for the provided node.

        :rype: ``list`` of :class:`GridscaleIp`
        """
        ips = self.ex_list_ips()

        result = []
        for ip in ips:
            related_servers = ip.extra.get("relations", {}).get("servers", [])
            for server in related_servers:
                # TODO: This is not consistent with volumes where key is
                # called "object_uuid"
                if server["server_uuid"] == node.id:
                    result.append(ip)

        return result

    def _to_node(self, data):
        extra_keys = ["cores", "power", "memory", "current_price", "relations"]

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)
        ips = []

        for diction in data["relations"]["public_ips"]:
            ips.append(diction["ip"])

        state = ""

        if data["power"] is True:
            state = NodeState.RUNNING
        else:
            state = NodeState.STOPPED

        node = Node(
            id=data["object_uuid"],
            name=data["name"],
            state=state,
            public_ips=ips,
            created_at=parse_date(data["create_time"]),
            private_ips=None,
            driver=self.connection.driver,
            extra=extra,
        )

        return node

    def _to_volume(self, data):
        extra_keys = ["create_time", "current_price", "storage_type", "relations"]

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        storage = StorageVolume(
            id=data["object_uuid"],
            name=data["name"],
            size=data["capacity"],
            driver=self.connection.driver,
            extra=extra,
        )

        return storage

    def _to_volume_snapshot(self, data):
        extra_keys = [
            "labels",
            "status",
            "usage_in_minutes",
            "location_country",
            "current_price",
            "parent_uuid",
        ]

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        volume_snapshot = VolumeSnapshot(
            id=data["object_uuid"],
            driver=self.connection.driver,
            size=data["capacity"],
            extra=extra,
            created=parse_date(data["create_time"]),
            state=StorageVolumeState.AVAILABLE,
            name=data["name"],
        )

        return volume_snapshot

    def _to_location(self, data):
        location = NodeLocation(
            id=data["object_uuid"],
            name=data["name"],
            country=data["country"],
            driver=self.connection.driver,
        )
        return location

    def _to_ip(self, data):
        extra_keys = [
            "create_time",
            "current_price",
            "name",
            "relations",
            "reverse_dns",
            "status",
        ]
        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        ip = GridscaleIp(
            id=data["object_uuid"],
            family=data["family"],
            prefix=data["prefix"],
            create_time=data["create_time"],
            address=data["ip"],
            extra=extra,
        )

        return ip

    def _to_network(self, data):
        network = GridscaleNetwork(
            id=data["object_uuid"],
            name=data["name"],
            create_time=data["create_time"],
            status=data["status"],
            relations=data["relations"],
        )
        return network

    def _to_node_image(self, data):
        extra_keys = [
            "capacity",
            "create_time",
            "labels",
            "ostype",
            "location_name",
            "private",
            "status",
            "usage_in_minutes",
            "version",
        ]

        extra = self._extract_values_to_dict(data=data, keys=extra_keys)

        template = NodeImage(
            id=data["object_uuid"],
            name=data["name"],
            driver=self.connection.driver,
            extra=extra,
        )

        return template

    def _to_key(self, data, name=None, sshkey=None):
        extra = {"uuid": data["object_uuid"], "labels": data.get("labels", [])}

        name = data.get("name", name)
        sshkey = data.get("sshkey", sshkey)

        key = KeyPair(
            name=name,
            fingerprint=data["object_uuid"],
            public_key=sshkey,
            private_key=None,
            extra=extra,
            driver=self.connection.driver,
        )

        return key

    def _extract_values_to_dict(self, data, keys):
        """
        Extract extra values to dict.

        :param data: dict to extract values from.
        :type data: ``dict``
        :param keys: keys to extract
        :type keys: ``List``
        :return: dictionary containing extra values
        :rtype: ``dict``
        """

        result = {}

        for key in keys:
            if key == "memory":
                result[key] = data[key] * 1024
            else:
                result[key] = data[key]

        return result

    def _get_response_dict(self, raw_response):
        """

        Get the actual response dictionary.

        :param raw_response: Nested dictionary.
        :type raw_response: ``dict``

        :return: Not-nested dictionary.
        :rtype: ``dict``
        """
        return list(raw_response.object.values())[0]

    def _get_resource(self, endpoint_suffix, object_uuid):
        """
        Get specific uuid specific resource.

        :param endpoint_suffix: Endpoint resource e.g. servers/.
        :type endpoint_suffix: ``str``
        :param object_uuid: Uuid of resource to be pulled.
        :type object_uuid: ``str``
        :return: Response dictionary.
        :rtype: nested ``dict``
        """
        data = self._sync_request(endpoint="objects/{}/{}".format(endpoint_suffix, object_uuid))

        data = self._get_response_dict(data)

        return data



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import json

from libcloud.test import MockHttp, LibcloudTestCase, unittest
from libcloud.compute import providers
from libcloud.utils.py3 import httplib
from libcloud.compute.base import NodeImage, NodeLocation, NodeAuthSSHKey
from libcloud.test.secrets import KAMATERA_PARAMS
from libcloud.compute.types import Provider, NodeState
from libcloud.common.exceptions import BaseHTTPError
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.kamatera import KamateraNodeDriver


class KamateraAuthenticationTests(LibcloudTestCase):
    def setUp(self):
        KamateraNodeDriver.connectionCls.conn_class = KamateraMockHttp
        self.driver = KamateraNodeDriver("nosuchuser", "nopwd")

    def test_authentication_fails(self):
        with self.assertRaises(BaseHTTPError):
            self.driver.list_locations()


class KamateraNodeDriverTests(LibcloudTestCase):
    def setUp(self):
        KamateraTestDriver.connectionCls.conn_class = KamateraMockHttp
        self.driver = KamateraTestDriver(*KAMATERA_PARAMS)
        self.eu_node_location = NodeLocation(
            id="EU", name="Amsterdam", country="The Netherlands", driver=self.driver
        )
        self.il_node_location = NodeLocation(
            id="IL", name="Rosh Haayin", country="Israel", driver=self.driver
        )
        self.centos_8_EU_node_image = NodeImage(
            id="EU:6000C2987c9641fd2619a149ba2ca01a",
            name="CentOS 8.0 64-bit - Minimal Configuration",
            driver=self.driver,
            extra={
                "datacenter": "EU",
                "os": "CentOS",
                "code": "8.0 64bit_minimal",
                "osDiskSizeGB": 5,
                "ramMBMin": {"A": 256, "B": 256, "T": 256, "D": 256},
            },
        )
        self.small_node_size = self.driver.ex_get_size(
            ramMB=4096,
            diskSizeGB=30,
            cpuType="B",
            cpuCores=2,
            monthlyTrafficPackage="t5000",
            id="small",
            name="small",
        )

    def test_creating_driver(self):
        cls = providers.get_driver(Provider.KAMATERA)
        self.assertIs(cls, KamateraNodeDriver)

    def test_features(self):
        features = self.driver.features["create_node"]
        self.assertIn("password", features)
        self.assertIn("generates_password", features)
        self.assertIn("ssh_key", features)

    def test_list_locations(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) == 13)
        self.assert_object(self.il_node_location, objects=locations)

    def test_list_sizes(self):
        sizes = self.driver.list_sizes(self.eu_node_location)
        self.assertTrue(len(sizes) >= 1)
        self.assert_object(self.small_node_size, objects=sizes)

    def test_list_images(self):
        images = self.driver.list_images(self.eu_node_location)
        self.assertTrue(len(images) > 10)
        self.assert_object(self.centos_8_EU_node_image, objects=images)

    def test_ex_list_capabilities(self):
        capabilities = self.driver.ex_list_capabilities(self.eu_node_location)
        self.assertEqual(
            {
                "cpuTypes",
                "defaultMonthlyTrafficPackage",
                "diskSizeGB",
                "monthlyTrafficPackage",
            },
            set(capabilities.keys()),
        )
        self.assertTrue(len(capabilities["cpuTypes"]), 4)
        self.assertEqual(
            {"id", "description", "name", "ramMB", "cpuCores"},
            set(capabilities["cpuTypes"][0]),
        )

    def test_create_node(self):
        node = self.driver.create_node(
            name="test_server",
            size=self.small_node_size,
            image=self.centos_8_EU_node_image,
            location=self.eu_node_location,
        )

        self.assertTrue(len(node.id) > 8)
        self.assertEqual(node.name, "my-server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)
        self.assertTrue(len(node.extra["generated_password"]) > 0)

    def test_create_node_with_ssh_keys(self):
        node = self.driver.create_node(
            name="test_server_pubkey",
            size=self.small_node_size,
            image=self.centos_8_EU_node_image,
            location=self.eu_node_location,
            auth=NodeAuthSSHKey("publickey"),
        )

        self.assertTrue(len(node.id) > 8)
        self.assertEqual(node.name, "my-server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)
        self.assertFalse("generated_password" in node.extra)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()
        self.assertTrue(len(nodes) >= 1)
        node = nodes[0]
        self.assertEqual(node.name, "test_server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertEqual(node.driver, self.driver)

    def test_list_nodes_full(self):
        nodes = self.driver.list_nodes(ex_full_details=True)
        self.assertTrue(len(nodes) >= 1)
        node = nodes[0]
        self.assertEqual(node.name, "my-server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)

    def test_reboot_node(self):
        nodes = self.driver.list_nodes()
        success = self.driver.reboot_node(nodes[0])
        self.assertTrue(success)

    def assert_object(self, expected_object, objects):
        same_data = any([self.objects_equals(expected_object, obj) for obj in objects])
        self.assertTrue(
            same_data,
            "Objects does not match ({}, {})".format(expected_object, objects[:2]),
        )

    def objects_equals(self, expected_obj, obj):
        for name in vars(expected_obj):
            expected_data = getattr(expected_obj, name)
            actual_data = getattr(obj, name)
            same_data = self.data_equals(expected_data, actual_data)
            if not same_data:
                break
        return same_data

    def data_equals(self, expected_data, actual_data):
        if isinstance(expected_data, dict):
            return self.dicts_equals(expected_data, actual_data)
        else:
            return expected_data == actual_data

    def dicts_equals(self, d1, d2):
        dict_keys_same = set(d1.keys()) == set(d2.keys())
        if not dict_keys_same:
            return False

        for key in d1.keys():
            if d1[key] != d2[key]:
                return False

        return True


class KamateraTestDriver(KamateraNodeDriver):
    def ex_wait_command(self, *args, **kwargs):
        kwargs["poll_interval_seconds"] = 0
        return KamateraNodeDriver.ex_wait_command(self, *args, **kwargs)


class KamateraMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("kamatera")

    def _service_server(self, method, url, body, headers):
        client_id, secret = headers["AuthClientId"], headers["AuthSecret"]
        if client_id == "nosuchuser" and secret == "nopwd":
            body = self.fixtures.load("failed_auth.json")
            status = httplib.UNAUTHORIZED
        else:
            if url == "/service/server" and json.loads(body).get("ssh-key"):
                body = self.fixtures.load("create_server_sshkey.json")
            else:
                body = self.fixtures.load(
                    {
                        "/service/server?datacenter=1": "datacenters.json",
                        "/service/server?sizes=1&datacenter=EU": "sizes_datacenter_EU.json",
                        "/service/server?images=1&datacenter=EU": "images_datacenter_EU.json",
                        "/service/server?capabilities=1&datacenter=EU": "capabilities_datacenter_EU.json",
                        "/service/server": "create_server.json",
                    }[url]
                )
            status = httplib.OK
        return status, body, {}, httplib.responses[status]

    def _service_queue(self, method, url, body, headers):
        if not hasattr(self, "_service_queue_call_count"):
            self._service_queue_call_count = 0
        self._service_queue_call_count += 1
        body = self.fixtures.load(
            {"/service/queue?id=12345": "queue_12345-%s.json" % self._service_queue_call_count}[url]
        )
        status = httplib.OK
        return status, body, {}, httplib.responses[status]

    def _service_server_info(self, method, url, body, headers):
        body = self.fixtures.load({"/service/server/info": "server_info.json"}[url])
        status = httplib.OK
        return status, body, {}, httplib.responses[status]

    def _service_servers(self, method, url, body, headers):
        body = self.fixtures.load({"/service/servers": "servers.json"}[url])
        status = httplib.OK
        return status, body, {}, httplib.responses[status]

    def _service_server_reboot(self, method, url, body, headers):
        body = self.fixtures.load({"/service/server/reboot": "server_operation.json"}[url])
        status = httplib.OK
        return status, body, {}, httplib.responses[status]


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.base import Zone, Record
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
)
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_LUADNS
from libcloud.dns.drivers.luadns import LuadnsDNSDriver
from libcloud.test.file_fixtures import DNSFileFixtures


class LuadnsTests(unittest.TestCase):
    def setUp(self):
        LuadnsMockHttp.type = None
        LuadnsDNSDriver.connectionCls.conn_class = LuadnsMockHttp
        self.driver = LuadnsDNSDriver(*DNS_PARAMS_LUADNS)
        self.test_zone = Zone(
            id="11",
            type="master",
            ttl=None,
            domain="example.com",
            extra={},
            driver=self.driver,
        )
        self.test_record = Record(
            id="13",
            type=RecordType.A,
            name="example.com",
            zone=self.test_zone,
            data="127.0.0.1",
            driver=self,
            extra={},
        )

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' % key)

    def test_list_zones_empty(self):
        LuadnsMockHttp.type = "EMPTY_ZONES_LIST"
        zones = self.driver.list_zones()

        self.assertEqual(zones, [])

    def test_list_zones_success(self):
        zones = self.driver.list_zones()

        self.assertEqual(len(zones), 2)

        zone = zones[0]
        self.assertEqual(zone.id, "1")
        self.assertEqual(zone.domain, "example.com")
        self.assertIsNone(zone.type)
        self.assertEqual(zone.driver, self.driver)
        self.assertIsNone(zone.ttl)

        second_zone = zones[1]
        self.assertEqual(second_zone.id, "2")
        self.assertEqual(second_zone.domain, "example.net")
        self.assertIsNone(second_zone.type)
        self.assertEqual(second_zone.driver, self.driver)
        self.assertIsNone(second_zone.ttl)

    def test_get_zone_zone_does_not_exist(self):
        LuadnsMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.get_zone(zone_id="13")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "13")
        else:
            self.fail("Exception was not thrown")

    def test_get_zone_success(self):
        LuadnsMockHttp.type = "GET_ZONE_SUCCESS"
        zone = self.driver.get_zone(zone_id="31")

        self.assertEqual(zone.id, "31")
        self.assertEqual(zone.domain, "example.org")
        self.assertIsNone(zone.type)
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

    def test_delete_zone_success(self):
        LuadnsMockHttp.type = "DELETE_ZONE_SUCCESS"
        zone = self.test_zone
        status = self.driver.delete_zone(zone=zone)

        self.assertEqual(status, True)

    def test_delete_zone_zone_does_not_exist(self):
        LuadnsMockHttp.type = "DELETE_ZONE_ZONE_DOES_NOT_EXIST"
        zone = self.test_zone
        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "11")
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_success(self):
        LuadnsMockHttp.type = "CREATE_ZONE_SUCCESS"
        zone = self.driver.create_zone(domain="example.org")

        self.assertEqual(zone.id, "3")
        self.assertEqual(zone.domain, "example.org")
        self.assertIsNone(zone.type)
        self.assertIsNone(zone.ttl)
        self.assertEqual(zone.driver, self.driver)

    def test_create_zone_zone_zone_already_exists(self):
        LuadnsMockHttp.type = "CREATE_ZONE_ZONE_ALREADY_EXISTS"
        try:
            self.driver.create_zone(domain="test.com")
        except ZoneAlreadyExistsError as e:
            self.assertEqual(e.zone_id, "test.com")
        else:
            self.fail("Exception was not thrown")

    def test_list_records_empty(self):
        LuadnsMockHttp.type = "EMPTY_RECORDS_LIST"
        zone = self.test_zone
        records = self.driver.list_records(zone=zone)

        self.assertEqual(records, [])

    def test_list_records_success(self):
        LuadnsMockHttp.type = "LIST_RECORDS_SUCCESS"
        zone = self.test_zone
        records = self.driver.list_records(zone=zone)

        self.assertEqual(len(records), 2)

        record = records[0]
        self.assertEqual(record.id, "6683")
        self.assertEqual(record.type, "NS")
        self.assertEqual(record.name, "example.org.")
        self.assertEqual(record.data, "b.ns.luadns.net.")
        self.assertEqual(record.zone, self.test_zone)
        self.assertEqual(record.zone.id, "11")

        second_record = records[1]
        self.assertEqual(second_record.id, "6684")
        self.assertEqual(second_record.type, "NS")
        self.assertEqual(second_record.name, "example.org.")
        self.assertEqual(second_record.data, "a.ns.luadns.net.")
        self.assertEqual(second_record.zone, self.test_zone)

    def test_get_record_record_does_not_exist(self):
        LuadnsMockHttp.type = "GET_RECORD_RECORD_DOES_NOT_EXIST"
        try:
            self.driver.get_record(zone_id="31", record_id="31")
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, "31")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_success(self):
        LuadnsMockHttp.type = "GET_RECORD_SUCCESS"
        record = self.driver.get_record(zone_id="31", record_id="31")

        self.assertEqual(record.id, "31")
        self.assertEqual(record.type, "MX")
        self.assertEqual(record.name, "example.com.")
        self.assertEqual(record.data, "10 mail.example.com.")

    def test_delete_record_success(self):
        LuadnsMockHttp.type = "DELETE_RECORD_SUCCESS"
        record = self.test_record
        status = self.driver.delete_record(record=record)

        self.assertEqual(status, True)

    def test_delete_record_RECORD_DOES_NOT_EXIST_ERROR(self):
        LuadnsMockHttp.type = "DELETE_RECORD_RECORD_DOES_NOT_EXIST"
        record = self.test_record
        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, "13")
        else:
            self.fail("Exception was not thrown")

    def test_create_record_success(self):
        LuadnsMockHttp.type = "CREATE_RECORD_SUCCESS"
        record = self.driver.create_record(
            name="test.com.",
            zone=self.test_zone,
            type="A",
            data="127.0.0.1",
            extra={"ttl": 13},
        )
        self.assertEqual(record.id, "31")
        self.assertEqual(record.name, "test.com.")
        self.assertEqual(record.data, "127.0.0.1")
        self.assertIsNone(record.ttl)

    def test_record_already_exists_error(self):
        pass


class LuadnsMockHttp(MockHttp):
    fixtures = DNSFileFixtures("luadns")

    def _v1_zones(self, method, url, body, headers):
        body = self.fixtures.load("zones_list.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_EMPTY_ZONES_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_zones_list.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_13_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_DELETE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_DELETE_ZONE_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("zone_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_CREATE_ZONE_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_CREATE_ZONE_ZONE_ALREADY_EXISTS(self, method, url, body, headers):
        body = self.fixtures.load("zone_already_exists.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_EMPTY_RECORDS_LIST(self, method, url, body, headers):
        body = self.fixtures.load("empty_records_list.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_LIST_RECORDS_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("records_list.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_records_31_GET_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("record_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_RECORD_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_GET_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_31_records_31_GET_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("get_record.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_13_DELETE_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("delete_record_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_13_DELETE_RECORD_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("record_does_not_exist.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]

    def _v1_zones_11_records_CREATE_RECORD_SUCCESS(self, method, url, body, headers):
        body = self.fixtures.load("create_record_success.json")

        return httplib.OK, body, {}, httplib.responses[httplib.OK]


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import hashlib
import datetime
from typing import Any, Dict, List, Union, Optional
from collections import OrderedDict

from libcloud.compute.base import Node, NodeSize, NodeImage
from libcloud.compute.types import NodeState
from libcloud.container.base import Container, ContainerImage, ContainerDriver, ContainerCluster
from libcloud.container.types import ContainerState
from libcloud.common.exceptions import BaseHTTPError
from libcloud.common.kubernetes import (
    KubernetesException,
    KubernetesDriverMixin,
    KubernetesBasicAuthConnection,
)
from libcloud.container.providers import Provider

__all__ = [
    "KubernetesContainerDriver",
    "to_n_bytes",
    "to_memory_str",
    "to_cpu_str",
    "to_n_cpus",
]


ROOT_URL = "/api/"


K8S_UNIT_MAP = OrderedDict(
    {
        "K": 1000,
        "Ki": 1024,
        "M": 1000 * 1000,
        "Mi": 1024 * 1024,
        "G": 1000 * 1000 * 1000,
        "Gi": 1024 * 1024 * 1024,
    }
)


def to_n_bytes(memory_str: str) -> int:
    """Convert memory string to number of bytes
    (e.g. '1234Mi'-> 1293942784)
    """
    if memory_str.startswith("0"):
        return 0
    if memory_str.isnumeric():
        return int(memory_str)
    for unit, multiplier in K8S_UNIT_MAP.items():
        if memory_str.endswith(unit):
            return int(memory_str.strip(unit)) * multiplier


def to_memory_str(n_bytes: int, unit: Optional[str] = None) -> str:
    """Convert number of bytes to k8s memory string
    (e.g. 1293942784 -> '1234Mi')
    """
    if n_bytes == 0:
        return "0K"
    n_bytes = int(n_bytes)
    memory_str = None
    if unit is None:
        for unit, multiplier in reversed(K8S_UNIT_MAP.items()):
            converted_n_bytes_float = n_bytes / multiplier
            converted_n_bytes = n_bytes // multiplier
            memory_str = f"{converted_n_bytes}{unit}"
            if converted_n_bytes_float % 1 == 0:
                break
    elif K8S_UNIT_MAP.get(unit):
        memory_str = f"{n_bytes // K8S_UNIT_MAP[unit]}{unit}"
    return memory_str


def to_cpu_str(n_cpus: Union[int, float]) -> str:
    """Convert number of cpus to cpu string
    (e.g. 0.5 -> '500m')
    """
    if n_cpus == 0:
        return "0"
    millicores = n_cpus * 1000
    if millicores % 1 == 0:
        return f"{int(millicores)}m"
    microcores = n_cpus * 1000000
    if microcores % 1 == 0:
        return f"{int(microcores)}u"
    nanocores = n_cpus * 1000000000
    return f"{int(nanocores)}n"


def to_n_cpus(cpu_str: str) -> Union[int, float]:
    """Convert cpu string to number of cpus
    (e.g. '500m' -> 0.5, '2000000000n' -> 2)
    """
    if cpu_str.endswith("n"):
        return int(cpu_str.strip("n")) / 1000000000
    elif cpu_str.endswith("u"):
        return int(cpu_str.strip("u")) / 1000000
    elif cpu_str.endswith("m"):
        return int(cpu_str.strip("m")) / 1000
    elif cpu_str.isnumeric():
        return int(cpu_str)
    else:
        return 0


def sum_resources(*resource_dicts):
    total_cpu = 0
    total_memory = 0
    for rd in resource_dicts:
        total_cpu += to_n_cpus(rd.get("cpu", "0m"))
        total_memory += to_n_bytes(rd.get("memory", "0K"))
    return {"cpu": to_cpu_str(total_cpu), "memory": to_memory_str(total_memory)}


class KubernetesDeployment:
    def __init__(
        self,
        id: str,
        name: str,
        namespace: str,
        created_at: str,
        replicas: int,
        selector: Dict[str, Any],
        extra: Optional[Dict[str, Any]] = None,
    ):
        self.id = id
        self.name = name
        self.namespace = namespace
        self.created_at = created_at
        self.replicas = replicas
        self.selector = selector
        self.extra = extra or {}

    def __repr__(self):
        return "<KubernetesDeployment name={} namespace={} replicas={}>".format(
            self.name,
            self.namespace,
            self.replicas,
        )


class KubernetesPod:
    def __init__(
        self,
        id: str,
        name: str,
        containers: List[Container],
        namespace: str,
        state: str,
        ip_addresses: List[str],
        created_at: datetime.datetime,
        node_name: str,
        extra: Dict[str, Any],
    ):
        """
        A Kubernetes pod
        """
        self.id = id
        self.name = name
        self.containers = containers
        self.namespace = namespace
        self.state = state
        self.ip_addresses = ip_addresses
        self.created_at = created_at
        self.node_name = node_name
        self.extra = extra

    def __repr__(self):
        return "<KubernetesPod name={} namespace={} state={}>".format(
            self.name,
            self.namespace,
            self.state,
        )


class KubernetesNamespace(ContainerCluster):
    """
    A Kubernetes namespace
    """

    def __repr__(self):
        return "<KubernetesNamespace name={}>".format(self.name)


class KubernetesContainerDriver(KubernetesDriverMixin, ContainerDriver):
    type = Provider.KUBERNETES
    name = "Kubernetes"
    website = "http://kubernetes.io"
    connectionCls = KubernetesBasicAuthConnection
    supports_clusters = True

    def list_containers(self, image=None, all=True) -> List[Container]:
        """
        List the deployed container images

        :param image: Filter to containers with a certain image(unused)
        :type  image: :class:`libcloud.container.base.ContainerImage`

        :param all: Show all container (unused)
        :type  all: ``bool``

        :rtype: ``list`` of :class:`libcloud.container.base.Container`
        """
        try:
            result = self.connection.request(ROOT_URL + "v1/pods").object
        except Exception as exc:
            errno = getattr(exc, "errno", None)
            if errno == 111:
                raise KubernetesException(
                    errno,
                    "Make sure kube host is accessible" "and the API port is correct",
                )
            raise

        pods = [self._to_pod(value) for value in result["items"]]
        containers = []
        for pod in pods:
            containers.extend(pod.containers)
        return containers

    def get_container(self, id: str) -> Container:
        """
        Get a container by ID

        :param id: The ID of the container to get
        :type  id: ``str``

        :rtype: :class:`libcloud.container.base.Container`
        """
        containers = self.list_containers()
        match = [container for container in containers if container.id == id]
        return match[0]

    def list_namespaces(self) -> List[KubernetesNamespace]:
        """
        Get a list of namespaces that pods can be deployed into

        :rtype: ``list`` of :class:`.KubernetesNamespace`
        """
        try:
            result = self.connection.request(ROOT_URL + "v1/namespaces/").object
        except Exception as exc:
            errno = getattr(exc, "errno", None)
            if errno == 111:
                raise KubernetesException(
                    errno,
                    "Make sure kube host is accessible" "and the API port is correct",
                )
            raise

        namespaces = [self._to_namespace(value) for value in result["items"]]
        return namespaces

    def get_namespace(self, id: str) -> KubernetesNamespace:
        """
        Get a namespace by ID

        :param id: The ID of the namespace to get
        :type  id: ``str``

        :rtype: :class:`.KubernetesNamespace`
        """
        result = self.connection.request(ROOT_URL + "v1/namespaces/%s" % id).object

        return self._to_namespace(result)

    def delete_namespace(self, namespace: KubernetesNamespace) -> bool:
        """
        Delete a namespace

        :return: ``True`` if the destroy was successful, otherwise ``False``.
        :rtype: ``bool``
        """
        self.connection.request(
            ROOT_URL + "v1/namespaces/%s" % namespace.id, method="DELETE"
        ).object
        return True

    def create_namespace(self, name: str) -> KubernetesNamespace:
        """
        Create a namespace

        :param  name: The name of the namespace
        :type   name: ``str``

        :rtype: :class:`.KubernetesNamespace`
        """
        request = {"metadata": {"name": name}}
        result = self.connection.request(
            ROOT_URL + "v1/namespaces", method="POST", data=json.dumps(request)
        ).object
        return self._to_namespace(result)

    def deploy_container(
        self,
        name: str,
        image: ContainerImage,
        namespace: KubernetesNamespace = None,
        parameters: Optional[str] = None,
        start: Optional[bool] = True,
    ):
        """
        Deploy an installed container image.
        In kubernetes this deploys a single container Pod.
        https://cloud.google.com/container-engine/docs/pods/single-container

        :param name: The name of the new container
        :type  name: ``str``

        :param image: The container image to deploy
        :type  image: :class:`.ContainerImage`

        :param namespace: The namespace to deploy to, None is default
        :type  namespace: :class:`.KubernetesNamespace`

        :param parameters: Container Image parameters(unused)
        :type  parameters: ``str``

        :param start: Start the container on deployment(unused)
        :type  start: ``bool``

        :rtype: :class:`.Container`
        """
        if namespace is None:
            namespace = "default"
        else:
            namespace = namespace.id
        request = {
            "metadata": {"name": name},
            "spec": {"containers": [{"name": name, "image": image.name}]},
        }
        result = self.connection.request(
            ROOT_URL + "v1/namespaces/%s/pods" % namespace,
            method="POST",
            data=json.dumps(request),
        ).object
        return self._to_namespace(result)

    def destroy_container(self, container: Container) -> bool:
        """
        Destroy a deployed container. Because the containers are single
        container pods, this will delete the pod.

        :param container: The container to destroy
        :type  container: :class:`.Container`

        :rtype: ``bool``
        """
        return self.ex_destroy_pod(container.extra["namespace"], container.extra["pod"])

    def ex_list_pods(self, fetch_metrics: bool = False) -> List[KubernetesPod]:
        """
        List available Pods

        :param fetch_metrics: Fetch metrics for pods
        :type  fetch_metrics: ``bool``

        :rtype: ``list`` of :class:`.KubernetesPod`
        """
        result = self.connection.request(ROOT_URL + "v1/pods").object
        metrics = None
        if fetch_metrics:
            try:
                metrics = {
                    (
                        metric["metadata"]["name"],
                        metric["metadata"]["namespace"],
                    ): metric["containers"]
                    for metric in self.ex_list_pods_metrics()
                }
            except BaseHTTPError:
                # Metrics Server may not be installed
                pass

        return [self._to_pod(value, metrics=metrics) for value in result["items"]]

    def ex_destroy_pod(self, namespace: str, pod_name: str) -> bool:
        """
        Delete a pod and the containers within it.

        :param namespace: The pod's namespace
        :type  namespace: ``str``

        :param pod_name: Name of the pod to destroy
        :type  pod_name: ``str``

        :rtype: ``bool``
        """
        self.connection.request(
            ROOT_URL + "v1/namespaces/{}/pods/{}".format(namespace, pod_name),
            method="DELETE",
        ).object
        return True

    def ex_list_nodes(self) -> List[Node]:
        """
        List available Nodes

        :rtype: ``list`` of :class:`.Node`
        """
        result = self.connection.request(ROOT_URL + "v1/nodes").object
        return [self._to_node(node) for node in result["items"]]

    def ex_destroy_node(self, node_name: str) -> bool:
        """
        Destroy a node.

        :param node_name: Name of the node to destroy
        :type  node_name: ``str``

        :rtype: ``bool``
        """
        self.connection.request(ROOT_URL + f"v1/nodes/{node_name}", method="DELETE").object
        return True

    def ex_get_version(self) -> str:
        """Get Kubernetes version

        :rtype: ``str``
        """
        return self.connection.request("/version").object["gitVersion"]

    def ex_list_nodes_metrics(self) -> List[Dict[str, Any]]:
        """Get nodes metrics from Kubernetes Metrics Server

        :rtype: ``list`` of ``dict``
        """
        return self.connection.request("/apis/metrics.k8s.io/v1beta1/nodes").object["items"]

    def ex_list_pods_metrics(self) -> List[Dict[str, Any]]:
        """Get pods metrics from Kubernetes Metrics Server

        :rtype: ``list`` of ``dict``
        """
        return self.connection.request("/apis/metrics.k8s.io/v1beta1/pods").object["items"]

    def ex_list_services(self) -> List[Dict[str, Any]]:
        """Get cluster services

        :rtype: ``list`` of ``dict``
        """
        return self.connection.request(ROOT_URL + "v1/services").object["items"]

    def ex_list_deployments(self) -> List[KubernetesDeployment]:
        """Get cluster deployments

        :rtype: ``list`` of :class:`.KubernetesDeployment`
        """
        items = self.connection.request("/apis/apps/v1/deployments").object["items"]
        return [self._to_deployment(item) for item in items]

    def _to_deployment(self, data):
        id_ = data["metadata"]["uid"]
        name = data["metadata"]["name"]
        namespace = data["metadata"]["namespace"]
        created_at = data["metadata"]["creationTimestamp"]
        replicas = data["spec"]["replicas"]
        selector = data["spec"]["selector"]

        extra = {
            "labels": data["metadata"]["labels"],
            "strategy": data["spec"]["strategy"]["type"],
            "total_replicas": data["status"]["replicas"],
            "updated_replicas": data["status"]["updatedReplicas"],
            "ready_replicas": data["status"]["readyReplicas"],
            "available_replicas": data["status"]["availableReplicas"],
            "conditions": data["status"]["conditions"],
        }
        return KubernetesDeployment(
            id=id_,
            name=name,
            namespace=namespace,
            created_at=created_at,
            replicas=replicas,
            selector=selector,
            extra=extra,
        )

    def _to_node(self, data):
        """
        Convert an API node data object to a `Node` object
        """
        ID = data["metadata"]["uid"]
        name = data["metadata"]["name"]
        driver = self.connection.driver
        memory = data["status"].get("capacity", {}).get("memory", "0K")
        cpu = data["status"].get("capacity", {}).get("cpu", "1")
        if isinstance(cpu, str) and not cpu.isnumeric():
            cpu = to_n_cpus(cpu)
        image_name = data["status"]["nodeInfo"].get("osImage")
        image = NodeImage(image_name, image_name, driver)
        size_name = f"{cpu} vCPUs, {memory} Ram"
        size_id = hashlib.md5(size_name.encode("utf-8")).hexdigest()
        extra_size = {"cpus": cpu}
        size = NodeSize(
            id=size_id,
            name=size_name,
            ram=memory,
            disk=0,
            bandwidth=0,
            price=0,
            driver=driver,
            extra=extra_size,
        )
        extra = {"memory": memory, "cpu": cpu}
        extra["os"] = data["status"]["nodeInfo"].get("operatingSystem")
        extra["kubeletVersion"] = data["status"]["nodeInfo"]["kubeletVersion"]
        extra["provider_id"] = data.get("spec", {}).get("providerID")
        for condition in data["status"]["conditions"]:
            if condition["type"] == "Ready" and condition["status"] == "True":
                state = NodeState.RUNNING
                break
        else:
            state = NodeState.UNKNOWN
        public_ips, private_ips = [], []
        for address in data["status"]["addresses"]:
            if address["type"] == "InternalIP":
                private_ips.append(address["address"])
            elif address["type"] == "ExternalIP":
                public_ips.append(address["address"])
        created_at = datetime.datetime.strptime(
            data["metadata"]["creationTimestamp"], "%Y-%m-%dT%H:%M:%SZ"
        )
        return Node(
            id=ID,
            name=name,
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=driver,
            image=image,
            size=size,
            extra=extra,
            created_at=created_at,
        )

    def _to_pod(self, data, metrics=None):
        """
        Convert an API response to a Pod object
        """
        id_ = data["metadata"]["uid"]
        name = data["metadata"]["name"]
        namespace = data["metadata"]["namespace"]
        state = data["status"]["phase"].lower()
        node_name = data["spec"].get("nodeName")
        container_statuses = data["status"].get("containerStatuses", {})
        containers = []
        extra = {"resources": {}}
        if metrics:
            try:
                extra["metrics"] = metrics[name, namespace]
            except KeyError:
                pass

        # response contains the status of the containers in a separate field
        for container in data["spec"]["containers"]:
            if container_statuses:
                spec = list(filter(lambda i: i["name"] == container["name"], container_statuses))[0]
            else:
                spec = container_statuses
            container_obj = self._to_container(container, spec, data)
            # Calculate new resources
            resources = extra["resources"]
            container_resources = container_obj.extra.get("resources", {})
            resources["limits"] = sum_resources(
                resources.get("limits", {}), container_resources.get("limits", {})
            )
            extra["resources"]["requests"] = sum_resources(
                resources.get("requests", {}), container_resources.get("requests", {})
            )
            containers.append(container_obj)
        ip_addresses = [ip_dict["ip"] for ip_dict in data["status"].get("podIPs", [])]
        created_at = datetime.datetime.strptime(
            data["metadata"]["creationTimestamp"], "%Y-%m-%dT%H:%M:%SZ"
        )

        return KubernetesPod(
            id=id_,
            name=name,
            namespace=namespace,
            state=state,
            ip_addresses=ip_addresses,
            containers=containers,
            created_at=created_at,
            node_name=node_name,
            extra=extra,
        )

    def _to_container(self, data, container_status, pod_data):
        """
        Convert container in Container instances
        """
        state = container_status.get("state")
        created_at = None
        if state:
            started_at = list(state.values())[0].get("startedAt")
            if started_at:
                created_at = datetime.datetime.strptime(started_at, "%Y-%m-%dT%H:%M:%SZ")
        extra = {
            "pod": pod_data["metadata"]["name"],
            "namespace": pod_data["metadata"]["namespace"],
        }
        resources = data.get("resources")
        if resources:
            extra["resources"] = resources
        return Container(
            id=container_status.get("containerID") or data["name"],
            name=data["name"],
            image=ContainerImage(
                id=container_status.get("imageID") or data["image"],
                name=data["image"],
                path=None,
                version=None,
                driver=self.connection.driver,
            ),
            ip_addresses=None,
            state=(ContainerState.RUNNING if container_status else ContainerState.UNKNOWN),
            driver=self.connection.driver,
            created_at=created_at,
            extra=extra,
        )

    def _to_namespace(self, data):
        """
        Convert an API node data object to a `KubernetesNamespace` object
        """
        return KubernetesNamespace(
            id=data["metadata"]["name"],
            name=data["metadata"]["name"],
            driver=self.connection.driver,
            extra={"phase": data["status"]["phase"]},
        )


def ts_to_str(timestamp):
    """
    Return a timestamp as a nicely formatted datetime string.
    """
    date = datetime.datetime.fromtimestamp(timestamp)
    date_string = date.strftime("%d/%m/%Y %H:%M %Z")
    return date_string



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Helper class for loading large fixture data

import os

from libcloud.utils.py3 import u

FIXTURES_ROOT = {
    "common": "common/fixtures",
    "compute": "compute/fixtures",
    "storage": "storage/fixtures",
    "loadbalancer": "loadbalancer/fixtures",
    "dns": "dns/fixtures",
    "drs": "drs/fixtures",
    "backup": "backup/fixtures",
    "openstack": "compute/fixtures/openstack",
    "container": "container/fixtures",
}


class FileFixtures:
    def __init__(self, fixtures_type, sub_dir=""):
        script_dir = os.path.abspath(os.path.split(__file__)[0])
        self.root = os.path.join(script_dir, FIXTURES_ROOT[fixtures_type], sub_dir)

    def load(self, file):
        path = os.path.join(self.root, file)
        if os.path.exists(path):
            with open(path, encoding="utf-8") as fh:
                content = fh.read()
            return u(content)
        else:
            raise OSError(path)


class ComputeFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="compute", sub_dir=sub_dir)


class StorageFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="storage", sub_dir=sub_dir)


class LoadBalancerFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="loadbalancer", sub_dir=sub_dir)


class DNSFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="dns", sub_dir=sub_dir)


class DRSFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="drs", sub_dir=sub_dir)


class OpenStackFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="openstack", sub_dir=sub_dir)


class ContainerFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="container", sub_dir=sub_dir)


class BackupFileFixtures(FileFixtures):
    def __init__(self, sub_dir=""):
        super().__init__(fixtures_type="backup", sub_dir=sub_dir)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Google Connection and Authentication classes.

Information about setting up your Google OAUTH2 credentials:

For libcloud, there are two basic methods for authenticating to Google using
OAUTH2: Service Accounts and Client IDs for Installed Applications.

Both are initially set up from the Cloud Console Console -
https://cloud.google.com/console

Setting up Service Account authentication (note that you need the cryptography
package installed to use this):

- Go to the Console
- Go to your project and then to "APIs & auth" on the left
- Click on "Credentials"
- Click on "Create New Client ID..."
- Select "Service account" and click on "Create Client ID"
- Download the Private Key (should happen automatically).  The key you download
  is in JSON format.
- Move the .json file to a safe location.
- Optionally, you may choose to Generate a PKCS12 key from the Console.
  It needs to be converted to the PEM format.  Please note, the PKCS12 format
  is deprecated and may be removed in a future release.
  - Convert the key using OpenSSL (the default password is 'notasecret').
  - Move the .pem file to a safe location.
- To Authenticate, you will need to pass the Service Account's "Email
  address" in as the user_id and the path to the .pem file as the key.

Setting up Installed Application authentication:

- Go to the Console
- Go to your project and then to "APIs & auth" on the left
- Click on "Credentials"
- Select "Installed application" and "Other" then click on
  "Create Client ID"
- To Authenticate, pass in the "Client ID" as the user_id and the "Client
  secret" as the key
- The first time that you do this, the libcloud will give you a URL to
  visit.  Copy and paste the URL into a browser.
- When you go to the URL it will ask you to log in (if you aren't already)
  and ask you if you want to allow the project access to your account.
- Click on Accept and you will be given a code.
- Paste that code at the prompt given to you by the Google libcloud
  connection.
- At that point, a token & refresh token will be stored in your home
  directory and will be used for authentication.

Please remember to secure your keys and access tokens.
"""


import os
import sys
import time
import errno
import base64
import logging
import datetime
import urllib.parse
from typing import Optional
from http.server import HTTPServer, BaseHTTPRequestHandler

from libcloud.utils.py3 import b, httplib, urlparse, urlencode
from libcloud.common.base import BaseDriver, JsonResponse, PollingConnection, ConnectionUserAndKey
from libcloud.common.types import LibcloudError, ProviderError
from libcloud.utils.connection import get_response_object

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore


try:
    from cryptography import exceptions
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.hashes import SHA256
    from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
except ImportError:
    # The cryptography library is unavailable
    SHA256 = None  # type: ignore

UTC_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

LOG = logging.getLogger(__name__)


def _utcnow():
    """
    Mocked in libcloud.test.common.google.GoogleTestCase.
    """
    return datetime.datetime.utcnow()


def _utc_timestamp(datetime_obj):
    """
    Return string of datetime_obj in the UTC Timestamp Format
    """
    return datetime_obj.strftime(UTC_TIMESTAMP_FORMAT)


def _from_utc_timestamp(timestamp):
    """
    Return datetime obj where date and time are pulled from timestamp string.
    """
    return datetime.datetime.strptime(timestamp, UTC_TIMESTAMP_FORMAT)


def _get_gce_metadata(path="", retry_failed: Optional[bool] = None):
    try:
        url = "http://metadata/computeMetadata/v1/" + path.lstrip("/")
        headers = {"Metadata-Flavor": "Google"}
        response = get_response_object(url, headers=headers, retry_failed=retry_failed)
        return response.status, "", response.body
    except Exception as e:
        return -1, str(e), None


class GoogleAuthError(LibcloudError):
    """Generic Error class for various authentication errors."""

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return repr(self.value)


class GoogleBaseError(ProviderError):
    def __init__(self, value, http_code, code, driver=None):
        self.code = code
        super().__init__(value, http_code, driver)


class InvalidRequestError(GoogleBaseError):
    pass


class JsonParseError(GoogleBaseError):
    pass


class ResourceNotFoundError(GoogleBaseError):
    def __init__(self, value, http_code, code, driver=None):
        self.code = code
        if (
            isinstance(value, dict)
            and "message" in value
            and value["message"].count("/") == 1
            and value["message"].count("projects/") == 1
        ):
            value["message"] = (
                value["message"] + ". A missing project "
                "error may be an authentication issue. "
                "Please  ensure your auth credentials match "
                "your project. "
            )
        super().__init__(value, http_code, driver)


class QuotaExceededError(GoogleBaseError):
    pass


class ResourceExistsError(GoogleBaseError):
    pass


class ResourceInUseError(GoogleBaseError):
    pass


class GoogleResponse(JsonResponse):
    """
    Google Base Response class.
    """

    def success(self):
        """
        Determine if the request was successful.

        For the Google response class, tag all responses as successful and
        raise appropriate Exceptions from parse_body.

        :return: C{True}
        """
        return True

    def _get_error(self, body):
        """
        Get the error code and message from a JSON response.

        Return just the first error if there are multiple errors.

        :param  body: The body of the JSON response dictionary
        :type   body: ``dict``

        :return:  Tuple containing error code and message
        :rtype:   ``tuple`` of ``str`` or ``int``
        """
        if "errors" in body["error"]:
            err = body["error"]["errors"][0]
        else:
            err = body["error"]

        if "code" in err:
            code = err.get("code")
            message = err.get("message")
        else:
            code = None
            if "reason" in err:
                code = err.get("reason")
            message = body.get("error_description", err)

        return (code, message)

    def parse_body(self):
        """
        Parse the JSON response body, or raise exceptions as appropriate.

        :return:  JSON dictionary
        :rtype:   ``dict``
        """
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        json_error = False
        try:
            body = json.loads(self.body)
        except Exception:
            # If there is both a JSON parsing error and an unsuccessful http
            # response (like a 404), we want to raise the http error and not
            # the JSON one, so don't raise JsonParseError here.
            body = self.body
            json_error = True

        valid_http_codes = [
            httplib.OK,
            httplib.CREATED,
            httplib.ACCEPTED,
            httplib.CONFLICT,
        ]
        if self.status in valid_http_codes:
            if json_error:
                raise JsonParseError(body, self.status, None)
            elif "error" in body:
                (code, message) = self._get_error(body)
                if code == "QUOTA_EXCEEDED":
                    raise QuotaExceededError(message, self.status, code)
                elif code == "RESOURCE_ALREADY_EXISTS":
                    raise ResourceExistsError(message, self.status, code)
                elif code == "alreadyExists":
                    raise ResourceExistsError(message, self.status, code)
                elif code.startswith("RESOURCE_IN_USE"):
                    raise ResourceInUseError(message, self.status, code)
                else:
                    raise GoogleBaseError(message, self.status, code)
            else:
                return body

        elif self.status == httplib.NOT_FOUND:
            if (not json_error) and ("error" in body):
                (code, message) = self._get_error(body)
            else:
                message = body
                code = None
            raise ResourceNotFoundError(message, self.status, code)

        elif self.status == httplib.BAD_REQUEST:
            if (not json_error) and ("error" in body):
                (code, message) = self._get_error(body)
            else:
                message = body
                code = None
            raise InvalidRequestError(message, self.status, code)

        else:
            if (not json_error) and ("error" in body):
                (code, message) = self._get_error(body)
            else:
                message = body
                code = None
            raise GoogleBaseError(message, self.status, code)


class GoogleBaseDriver(BaseDriver):
    name = "Google API"


class GoogleBaseAuthConnection(ConnectionUserAndKey):
    """
    Base class for Google Authentication.  Should be subclassed for specific
    types of authentication.
    """

    driver = GoogleBaseDriver
    responseCls = GoogleResponse
    name = "Google Auth"
    host = "accounts.google.com"
    auth_path = "/o/oauth2/auth"
    redirect_uri_port = 8087

    def __init__(
        self,
        user_id,
        key=None,
        scopes=None,
        redirect_uri="http://127.0.0.1",
        login_hint=None,
        **kwargs,
    ):
        """
        :param  user_id: The email address (for service accounts) or Client ID
                         (for installed apps) to be used for authentication.
        :type   user_id: ``str``

        :param  key: The RSA Key (for service accounts) or file path containing
                     key or Client Secret (for installed apps) to be used for
                     authentication.
        :type   key: ``str``

        :param  scopes: A list of urls defining the scope of authentication
                       to grant.
        :type   scopes: ``list``

        :keyword  redirect_uri: The Redirect URI for the authentication
                                request.  See Google OAUTH2 documentation for
                                more info.
        :type     redirect_uri: ``str``

        :keyword  login_hint: Login hint for authentication request.  Useful
                              for Installed Application authentication.
        :type     login_hint: ``str``
        """
        scopes = scopes or []

        self.scopes = " ".join(scopes)
        self.redirect_uri = redirect_uri
        self.login_hint = login_hint

        super().__init__(user_id, key, **kwargs)

    def add_default_headers(self, headers):
        """
        Add defaults for 'Content-Type' and 'Host' headers.
        """
        headers["Content-Type"] = "application/x-www-form-urlencoded"
        headers["Host"] = self.host
        return headers

    def _token_request(self, request_body):
        """
        Return an updated token from a token request body.

        :param  request_body: A dictionary of values to send in the body of the
                              token request.
        :type   request_body: ``dict``

        :return:  A dictionary with updated token information
        :rtype:   ``dict``
        """
        data = urlencode(request_body)
        try:
            response = self.request("/o/oauth2/token", method="POST", data=data)
        except AttributeError:
            raise GoogleAuthError(
                "Invalid authorization response, please " "check your credentials and time drift."
            )
        token_info = response.object
        if "expires_in" in token_info:
            expire_time = _utcnow() + datetime.timedelta(seconds=token_info["expires_in"])
            token_info["expire_time"] = _utc_timestamp(expire_time)
        return token_info

    def refresh_token(self, token_info):
        """
        Refresh the current token.

        Fetch an updated refresh token from internal metadata service.

        :param  token_info: Dictionary containing token information.
                            (Not used, but here for compatibility)
        :type   token_info: ``dict``

        :return:  A dictionary containing updated token information.
        :rtype:   ``dict``
        """
        # pylint: disable=no-member
        return self.get_new_token()


class GoogleInstalledAppAuthConnection(GoogleBaseAuthConnection):
    """Authentication connection for "Installed Application" authentication."""

    _state = "Libcloud Request"

    def get_code(self):
        """
        Give the user a URL that they can visit to authenticate.

        Mocked in libcloud.test.common.google.GoogleTestCase.

        :return:  Code supplied by the user after authenticating
        :rtype:   ``str``
        """
        auth_params = {
            "response_type": "code",
            "client_id": self.user_id,
            "redirect_uri": self._redirect_uri_with_port,
            "scope": self.scopes,
            "state": self._state,
        }
        if self.login_hint:
            auth_params["login_hint"] = self.login_hint

        data = urlencode(auth_params)

        url = "https://{}{}?{}".format(self.host, self.auth_path, data)
        print("\nPlease Go to the following URL and sign in:")
        print(url)
        code = self._receive_code_through_local_loopback()
        return code

    def get_new_token(self):
        """
        Get a new token. Generally used when no previous token exists or there
        is no refresh token

        :return:  Dictionary containing token information
        :rtype:   ``dict``
        """
        code = self.get_code()

        token_request = {
            "code": code,
            "client_id": self.user_id,
            "client_secret": self.key,
            "redirect_uri": self._redirect_uri_with_port,
            "grant_type": "authorization_code",
        }

        return self._token_request(token_request)

    def refresh_token(self, token_info):
        """
        Use the refresh token supplied in the token info to get a new token.

        :param  token_info: Dictionary containing current token information
        :type   token_info: ``dict``

        :return:  A dictionary containing updated token information.
        :rtype:   ``dict``
        """
        if "refresh_token" not in token_info:
            return self.get_new_token()
        refresh_request = {
            "refresh_token": token_info["refresh_token"],
            "client_id": self.user_id,
            "client_secret": self.key,
            "grant_type": "refresh_token",
        }

        new_token = self._token_request(refresh_request)
        if "refresh_token" not in new_token:
            new_token["refresh_token"] = token_info["refresh_token"]
        return new_token

    @property
    def _redirect_uri_with_port(self):
        return self.redirect_uri + ":" + str(self.redirect_uri_port)

    def _receive_code_through_local_loopback(self):
        """
        Start a local HTTP server that listens to a single GET request that is expected to be made
        by the loopback in the sign-in process and stops again afterwards.
        See https://developers.google.com/identity/protocols/oauth2/native-app#redirect-uri_loopback

        :return: The access code that was extracted from the local loopback GET request
        :rtype: ``str``
        """
        access_code = None

        class AccessCodeReceiver(BaseHTTPRequestHandler):
            # noinspection PyMethodParameters,PyPep8Naming
            def do_GET(self_):  # pylint: disable=no-self-argument
                query = urlparse.urlparse(self_.path).query
                query_components = dict(qc.split("=") for qc in query.split("&"))
                if "state" in query_components and query_components["state"] != urllib.parse.quote(
                    self._state
                ):
                    raise ValueError(
                        "States do not match: {} != {}, can't trust authentication".format(
                            self._state, query_components["state"]
                        )
                    )
                nonlocal access_code
                access_code = query_components["code"]
                self_.send_response(200)
                self_.send_header("Content-type", "text/html")
                self_.end_headers()
                self_.wfile.write(b"<html><head><title>Libcloud Sign-In</title></head>")
                self_.wfile.write(b"<body><p>You can now close this tab</p>")

        if (
            "127.0.0.1" in self.redirect_uri
            or "[::1]" in self.redirect_uri
            or "localhost" in self.redirect_uri
        ):
            # HTTPServer does not understand localhost unless you explicitly call it so
            server_address = "localhost", self.redirect_uri_port
        else:
            server_address = self.redirect_uri, self.redirect_uri_port

        server = HTTPServer(server_address=server_address, RequestHandlerClass=AccessCodeReceiver)
        # Only waits for a single request and stops afterwards
        server.handle_request()
        if access_code is None:
            raise RuntimeError(
                "Could not receive OAuth2 code: could not extract code though loopback"
            )
        return access_code


class GoogleServiceAcctAuthConnection(GoogleBaseAuthConnection):
    """Authentication class for "Service Account" authentication."""

    def __init__(self, user_id, key, *args, **kwargs):
        """
        Check to see if cryptography is available, and convert PEM key file
        into a key string, or extract the key from JSON object, string or
        file.

        :param  user_id: Email address to be used for Service Account
                authentication.
        :type   user_id: ``str``

        :param  key: The path to a PEM/JSON file containing the private RSA
        key, or a str/dict containing the PEM/JSON.
        :type   key: ``str`` or ``dict``

        """
        if SHA256 is None:
            raise GoogleAuthError(
                "cryptography library required for " "Service Account Authentication."
            )

        if isinstance(key, dict):
            # if it's a dict, assume it's containing parsed JSON
            key_content = key
            key = None
        else:
            key_path = os.path.expanduser(key)
            if os.path.exists(key_path) and os.path.isfile(key_path):
                # Assume it's a file and read it
                try:
                    with open(key_path) as f:
                        key_content = f.read()
                except OSError:
                    raise GoogleAuthError("Missing (or unreadable) key " "file: '%s'" % key)
            else:
                # assume it's a PEM str or serialized JSON str
                key_content = key
                key = None

        try:
            key_content = json.loads(key_content)
            # check if serialized JSON given
        except TypeError:
            # already a dict
            pass
        except ValueError:
            # it's been a PEM string all along
            pass
        finally:
            if "private_key" in key_content:
                key = key_content["private_key"]
            else:
                key = key_content

        try:
            # check if the key is actually a PEM encoded private key
            serialization.load_pem_private_key(b(key), password=None, backend=default_backend())
        except ValueError as e:
            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)
        except TypeError as e:
            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)
        except exceptions.UnsupportedAlgorithm as e:
            raise GoogleAuthError("Unable to decode provided PEM key: %s" % e)

        super().__init__(user_id, key, *args, **kwargs)

    def get_new_token(self):
        """
        Get a new token using the email address and RSA Key.

        :return:  Dictionary containing token information
        :rtype:   ``dict``
        """
        # The header is always the same
        header = {"alg": "RS256", "typ": "JWT"}
        header_enc = base64.urlsafe_b64encode(b(json.dumps(header)))

        # Construct a claim set
        claim_set = {
            "iss": self.user_id,
            "scope": self.scopes,
            "aud": "https://accounts.google.com/o/oauth2/token",
            "exp": int(time.time()) + 3600,
            "iat": int(time.time()),
        }
        claim_set_enc = base64.urlsafe_b64encode(b(json.dumps(claim_set)))

        # The message contains both the header and claim set
        message = b".".join((header_enc, claim_set_enc))
        # Then the message is signed using the key supplied
        key = serialization.load_pem_private_key(
            b(self.key), password=None, backend=default_backend()
        )
        signature = key.sign(data=b(message), padding=PKCS1v15(), algorithm=SHA256())
        signature = base64.urlsafe_b64encode(signature)

        # Finally the message and signature are sent to get a token
        jwt = b".".join((message, signature))
        request = {
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "assertion": jwt,
        }

        return self._token_request(request)


class GoogleGCEServiceAcctAuthConnection(GoogleBaseAuthConnection):
    """Authentication class for self-authentication when used with a GCE
    instance that supports serviceAccounts.
    """

    def get_new_token(self):
        """
        Get a new token from the internal metadata service.

        :return:  Dictionary containing token information
        :rtype:   ``dict``
        """
        path = "/instance/service-accounts/default/token"
        http_code, http_reason, token_info = _get_gce_metadata(path)
        if http_code == httplib.NOT_FOUND:
            raise ValueError("Service Accounts are not enabled for this " "GCE instance.")
        if http_code != httplib.OK:
            raise ValueError("Internal GCE Authorization failed: " "'%s'" % str(http_reason))
        token_info = json.loads(token_info)
        if "expires_in" in token_info:
            expire_time = _utcnow() + datetime.timedelta(seconds=token_info["expires_in"])
            token_info["expire_time"] = _utc_timestamp(expire_time)
        return token_info


class GoogleAuthType:
    """
    SA (Service Account),
    IA (Installed Application),
    GCE (Auth from a GCE instance with service account enabled)
    GCS_S3 (Cloud Storage S3 interoperability authentication)
    """

    SA = "SA"
    IA = "IA"
    GCE = "GCE"
    GCS_S3 = "GCS_S3"

    ALL_TYPES = [SA, IA, GCE, GCS_S3]
    OAUTH2_TYPES = [SA, IA, GCE]

    @classmethod
    def guess_type(cls, user_id):
        if cls._is_sa(user_id):
            return cls.SA
        elif cls._is_gcs_s3(user_id):
            return cls.GCS_S3
        elif cls._is_installed_application(user_id):
            # NOTE: This should be before "_is_gce()" call so we avoid
            # querying GCE metadata service if that's not necessary
            return cls.IA
        elif cls._is_gce():
            return cls.GCE
        else:
            # TODO: It's probably safe to throw here, but we return cls.IA
            # for backward compatibility reasons
            return cls.IA

    @classmethod
    def is_oauth2(cls, auth_type):
        return auth_type in cls.OAUTH2_TYPES

    @staticmethod
    def _is_installed_application(user_id):
        return user_id.endswith("apps.googleusercontent.com")

    @staticmethod
    def _is_gce():
        """
        Checks if we can access the GCE metadata server.
        Mocked in libcloud.test.common.google.GoogleTestCase.
        """
        # When using oAuth credentials we check for metadata server first, so
        # if server is unavailable and we retry many times before timing out,
        # this will slow down the driver instantiation when retrying failed
        # requests is enabled globally.
        http_code, http_reason, body = _get_gce_metadata(retry_failed=False)
        if http_code == httplib.OK and body:
            return True
        return False

    @staticmethod
    def _is_gcs_s3(user_id):
        """
        Checks S3 key format: alphanumeric chars starting with GOOG.
        """
        return user_id.startswith("GOOG")

    @staticmethod
    def _is_sa(user_id):
        return user_id.endswith(".gserviceaccount.com")


class GoogleOAuth2Credential:
    default_credential_file = "~/.google_libcloud_auth"

    def __init__(self, user_id, key, auth_type=None, credential_file=None, scopes=None, **kwargs):
        self.auth_type = auth_type or GoogleAuthType.guess_type(user_id)
        if self.auth_type not in GoogleAuthType.ALL_TYPES:
            raise GoogleAuthError("Invalid auth type: %s" % self.auth_type)
        if not GoogleAuthType.is_oauth2(self.auth_type):
            raise GoogleAuthError("Auth type %s cannot be used with OAuth2" % self.auth_type)
        self.user_id = user_id
        self.key = key

        default_credential_file = ".".join([self.default_credential_file, user_id])
        self.credential_file = credential_file or default_credential_file
        # Default scopes to read/write for compute, storage, and dns.
        self.scopes = scopes or [
            "https://www.googleapis.com/auth/compute",
            "https://www.googleapis.com/auth/devstorage.full_control",
            "https://www.googleapis.com/auth/ndev.clouddns.readwrite",
        ]

        self.token = self._get_token_from_file()

        if self.auth_type == GoogleAuthType.GCE:
            self.oauth2_conn = GoogleGCEServiceAcctAuthConnection(
                self.user_id, self.scopes, **kwargs
            )
        elif self.auth_type == GoogleAuthType.SA:
            self.oauth2_conn = GoogleServiceAcctAuthConnection(
                self.user_id, self.key, self.scopes, **kwargs
            )
        elif self.auth_type == GoogleAuthType.IA:
            self.oauth2_conn = GoogleInstalledAppAuthConnection(
                self.user_id, self.key, self.scopes, **kwargs
            )
        else:
            raise GoogleAuthError("Invalid auth_type: %s" % str(self.auth_type))

        if self.token is None:
            self.token = self.oauth2_conn.get_new_token()
            self._write_token_to_file()

    @property
    def access_token(self):
        if self.token_expire_utc_datetime < _utcnow():
            self._refresh_token()
        return self.token["access_token"]

    @property
    def token_expire_utc_datetime(self):
        return _from_utc_timestamp(self.token["expire_time"])

    def _refresh_token(self):
        self.token = self.oauth2_conn.refresh_token(self.token)
        self._write_token_to_file()

    def _get_token_from_file(self):
        """
        Read credential file and return token information.
        Mocked in libcloud.test.common.google.GoogleTestCase.

        :return:  Token information dictionary, or None
        :rtype:   ``dict`` or ``None``
        """
        token = None
        filename = os.path.realpath(os.path.expanduser(self.credential_file))

        try:
            with open(filename) as f:
                data = f.read()
            token = json.loads(data)
        except (OSError, ValueError) as e:
            # Note: File related errors (IOError) and errors related to json
            # parsing of the data (ValueError) are not fatal.
            LOG.info('Failed to read cached auth token from file "%s": %s', filename, str(e))

        return token

    def _write_token_to_file(self):
        """
        Write token to credential file.
        Mocked in libcloud.test.common.google.GoogleTestCase.
        """
        filename = os.path.expanduser(self.credential_file)
        filename = os.path.realpath(filename)

        try:
            data = json.dumps(self.token)
            write_flags = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
            with os.fdopen(os.open(filename, write_flags, int("600", 8)), "w") as f:
                f.write(data)
        except Exception as e:
            # Note: Failure to write (cache) token in a file is not fatal. It
            # simply means degraded performance since we will need to acquire a
            # new token each time script runs.
            LOG.info('Failed to write auth token to file "%s": %s', filename, str(e))


class GoogleBaseConnection(ConnectionUserAndKey, PollingConnection):
    """Base connection class for interacting with Google APIs."""

    driver = GoogleBaseDriver
    responseCls = GoogleResponse
    host = "www.googleapis.com"
    poll_interval = 2.0
    timeout = 180

    def __init__(
        self,
        user_id,
        key=None,
        auth_type=None,
        credential_file=None,
        scopes=None,
        **kwargs,
    ):
        """
        Determine authentication type, set up appropriate authentication
        connection and get initial authentication information.

        :param  user_id: The email address (for service accounts) or Client ID
                         (for installed apps) to be used for authentication.
        :type   user_id: ``str``

        :param  key: The RSA Key (for service accounts) or file path containing
                     key or Client Secret (for installed apps) to be used for
                     authentication.
        :type   key: ``str``

        :keyword  auth_type: See GoogleAuthType class for list and description
                             of accepted values.
                             If not supplied, auth_type will be guessed based
                             on value of user_id or if the code is running
                             on a GCE instance.
        :type     auth_type: ``str``

        :keyword  credential_file: Path to file for caching authentication
                                   information.
        :type     credential_file: ``str``

        :keyword  scopes: List of OAuth2 scope URLs. The empty default sets
                          read/write access to Compute, Storage, and DNS.
        :type     scopes: ``list``
        """
        super().__init__(user_id, key, **kwargs)

        self.oauth2_credential = GoogleOAuth2Credential(
            user_id, key, auth_type, credential_file, scopes, **kwargs
        )

        python_ver = "{}.{}.{}".format(
            sys.version_info[0],
            sys.version_info[1],
            sys.version_info[2],
        )
        ver_platform = "Python {}/{}".format(python_ver, sys.platform)
        self.user_agent_append(ver_platform)

    def add_default_headers(self, headers):
        """
        @inherits: :class:`Connection.add_default_headers`
        """
        headers["Content-Type"] = "application/json"
        headers["Host"] = self.host
        return headers

    def pre_connect_hook(self, params, headers):
        """
        Check to make sure that token hasn't expired.  If it has, get an
        updated token.  Also, add the token to the headers.

        @inherits: :class:`Connection.pre_connect_hook`
        """
        headers["Authorization"] = "Bearer " + self.oauth2_credential.access_token
        return params, headers

    def encode_data(self, data):
        """Encode data to JSON"""
        return json.dumps(data)

    def request(self, *args, **kwargs):
        """
        @inherits: :class:`Connection.request`
        """
        # Adds some retry logic for the occasional
        # "Connection Reset by peer" error.
        retries = 4
        tries = 0
        while tries < (retries - 1):
            try:
                return super().request(*args, **kwargs)
            except OSError as e:
                if e.errno == errno.ECONNRESET:
                    tries = tries + 1
                else:
                    raise e
        # One more time, then give up.
        return super().request(*args, **kwargs)

    def has_completed(self, response):
        """
        Determine if operation has completed based on response.

        :param  response: JSON response
        :type   response: I{responseCls}

        :return:  True if complete, False otherwise
        :rtype:   ``bool``
        """
        if response.object["status"] == "DONE":
            return True
        else:
            return False

    def get_poll_request_kwargs(self, response, context, request_kwargs):
        """
        @inherits: :class:`PollingConnection.get_poll_request_kwargs`
        """
        return {"action": response.object["selfLink"]}

    def morph_action_hook(self, action):
        """
        Update action to correct request path.

        In many places, the Google API returns a full URL to a resource.
        This will strip the scheme and host off of the path and just return
        the request.  Otherwise, it will prepend the base request_path to
        the action.

        :param  action: The action to be called in the http request
        :type   action: ``str``

        :return:  The modified request based on the action
        :rtype:   ``str``
        """
        if action.startswith("https://"):
            u = urlparse.urlsplit(action)
            request = urlparse.urlunsplit(("", "", u[2], u[3], u[4]))
        else:
            request = self.request_path + action
        return request



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest

from libcloud.test import MockHttp, LibcloudTestCase
from libcloud.utils.py3 import httplib
from libcloud.common.types import ServiceUnavailableError
from libcloud.compute.base import NodeSize, NodeImage
from libcloud.test.secrets import VULTR_PARAMS
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.vultr import VultrNodeDriver, VultrNodeDriverV1

try:
    import simplejson as json  # pylint: disable=unused-import
except ImportError:
    # pylint: disable=unused-import
    import json  # NOQA


# class VultrTests(unittest.TestCase, TestCaseMixin):
class VultrTests(LibcloudTestCase):
    def setUp(self):
        VultrNodeDriver.connectionCls.conn_class = VultrMockHttp
        VultrMockHttp.type = None
        self.driver = VultrNodeDriver(*VULTR_PARAMS, api_version="1")

    def test_correct_class_is_used(self):
        self.assertIsInstance(self.driver, VultrNodeDriverV1)

    def test_list_images_dont_require_api_key(self):
        self.driver.list_images()
        self.assertFalse(self.driver.connection.require_api_key())

    def test_list_images_success(self):
        images = self.driver.list_images()
        self.assertTrue(len(images) >= 1)

        image = images[0]
        self.assertTrue(image.id is not None)
        self.assertTrue(image.name is not None)

    def test_list_sizes_dont_require_api_key(self):
        self.driver.list_sizes()
        self.assertFalse(self.driver.connection.require_api_key())

    def test_list_sizes_success(self):
        """count of current plans"""
        sizes = self.driver.list_sizes()
        self.assertTrue(len(sizes) == 19)

        size = sizes[0]
        self.assertTrue(size.id.isdigit())
        self.assertEqual(size.name, "8192 MB RAM,110 GB SSD,10.00 TB BW")
        self.assertEqual(size.ram, 8192)

        size = sizes[16]
        self.assertTrue(size.id.isdigit())
        self.assertEqual(size.name, "16384 MB RAM,384 GB SSD,5.00 TB BW")
        self.assertEqual(size.ram, 16384)

    def test_list_locations_dont_require_api_key(self):
        self.driver.list_locations()
        self.assertFalse(self.driver.connection.require_api_key())

    def test_list_locations_success(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) >= 1)

        location = locations[0]
        self.assertEqual(location.id, "1")
        self.assertEqual(location.name, "New Jersey")
        self.assertEqual(location.extra["continent"], "North America")

    def test_list_locations_extra_success(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) >= 1)
        extra_keys = [
            "continent",
            "state",
            "ddos_protection",
            "block_storage",
            "regioncode",
        ]
        for location in locations:
            self.assertTrue(len(location.extra.keys()) >= 5)
            self.assertTrue(all(item in location.extra.keys() for item in extra_keys))

    def test_list_nodes_require_api_key(self):
        self.driver.list_nodes()
        self.assertTrue(self.driver.connection.require_api_key())

    def test_list_nodes_success(self):
        nodes = self.driver.list_nodes()
        self.assertEqual(len(nodes), 3)
        self.assertTrue(nodes[0].id.isdigit())
        self.assertEqual(nodes[0].id, "41306569")
        self.assertEqual(nodes[0].public_ips, ["45.76.43.87"])
        self.assertEqual(nodes[0].private_ips, ["10.7.96.85"])
        self.assertEqual(nodes[2].private_ips, [])

    def test_list_nodes_image_success(self):
        nodes = self.driver.list_nodes()
        node = nodes[0]
        self.assertTrue(isinstance(node.image, NodeImage))

    def test_list_nodes_size_success(self):
        nodes = self.driver.list_nodes()
        node = nodes[0]
        self.assertTrue(isinstance(node.size, NodeSize))

    def test_list_nodes_success_extra(self):
        extra_keys = [
            "default_password",
            "pending_charges",
            "cost_per_month",
        ]
        nodes = self.driver.list_nodes()
        for node in nodes:
            self.assertTrue(len(node.extra.keys()) > 5)
            self.assertTrue(all(item in node.extra.keys() for item in extra_keys))

    def test_reboot_node_success(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.reboot_node(node)
        self.assertTrue(result)

    def test_create_node_success(self):
        test_size = self.driver.list_sizes()[0]
        test_image = self.driver.list_images()[0]
        test_location = self.driver.list_locations()[0]
        created_node = self.driver.create_node("test-node", test_size, test_image, test_location)
        self.assertEqual(created_node.id, "41326859")

    def test_destroy_node_success(self):
        node = self.driver.list_nodes()[0]
        result = self.driver.destroy_node(node)
        self.assertTrue(result)

    def test_list_key_pairs_success(self):
        key_pairs = self.driver.list_key_pairs()
        self.assertEqual(len(key_pairs), 1)
        key_pair = key_pairs[0]
        self.assertEqual(key_pair.id, "5806a8ef2a0c6")
        self.assertEqual(key_pair.name, "test-key-pair")

    def test_create_key_pair_success(self):
        res = self.driver.create_key_pair("test-key-pair")
        self.assertTrue(res)

    def test_delete_key_pair_success(self):
        key_pairs = self.driver.list_key_pairs()
        key_pair = key_pairs[0]
        res = self.driver.delete_key_pair(key_pair)
        self.assertTrue(res)

    def test_rate_limit(self):
        VultrMockHttp.type = "SERVICE_UNAVAILABLE"
        self.assertRaises(ServiceUnavailableError, self.driver.list_nodes)


class VultrMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("vultr")

    # pylint: disable=unused-argument
    def _v1_regions_list(self, method, url, body, headers):
        body = self.fixtures.load("list_locations.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_os_list(self, method, url, body, headers):
        body = self.fixtures.load("list_images.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_plans_list(self, method, url, body, headers):
        body = self.fixtures.load("list_sizes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_server_list(self, method, url, body, headers):
        body = self.fixtures.load("list_nodes.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_server_list_SERVICE_UNAVAILABLE(self, method, url, body, headers):
        body = self.fixtures.load("error_rate_limit.txt")
        return (
            httplib.SERVICE_UNAVAILABLE,
            body,
            {},
            httplib.responses[httplib.SERVICE_UNAVAILABLE],
        )

    # pylint: disable=unused-argument
    def _v1_server_create(self, method, url, body, headers):
        body = self.fixtures.load("create_node.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_server_destroy(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_server_reboot(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_sshkey_list(self, method, url, body, headers):
        body = self.fixtures.load("list_key_pairs.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_sshkey_create(self, method, url, body, headers):
        body = self.fixtures.load("create_key_pair.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    # pylint: disable=unused-argument
    def _v1_sshkey_destroy(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Driver for Backblaze B2 service.
"""

import base64
import hashlib

from libcloud.utils.py3 import b, next, httplib, urlparse
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.utils.files import read_in_chunks, exhaust_iterator
from libcloud.common.types import LibcloudError, InvalidCredsError
from libcloud.storage.base import Object, Container, StorageDriver
from libcloud.utils.escape import sanitize_object_name
from libcloud.storage.types import ObjectDoesNotExistError, ContainerDoesNotExistError
from libcloud.storage.providers import Provider

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore


__all__ = [
    "BackblazeB2StorageDriver",
    "BackblazeB2Connection",
    "BackblazeB2AuthConnection",
]

AUTH_API_HOST = "api.backblaze.com"
API_PATH = "/b2api/v1/"


class BackblazeB2Response(JsonResponse):
    def success(self):
        return self.status in [httplib.OK, httplib.CREATED, httplib.ACCEPTED]

    def parse_error(self):
        status = int(self.status)
        body = self.parse_body()

        if status == httplib.UNAUTHORIZED:
            raise InvalidCredsError(body["message"])

        return self.body


class BackblazeB2AuthConnection(ConnectionUserAndKey):
    host = AUTH_API_HOST
    secure = True
    responseCls = BackblazeB2Response

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Those attributes are populated after authentication
        self.account_id = None
        self.api_url = None
        self.api_host = None
        self.download_url = None
        self.download_host = None
        self.auth_token = None

    def authenticate(self, force=False):
        """
        :param force: Force authentication if if we have already obtained the
                      token.
        :type force: ``bool``
        """
        if not self._is_authentication_needed(force=force):
            return self

        headers = {}
        action = "b2_authorize_account"
        auth_b64 = base64.b64encode(b("{}:{}".format(self.user_id, self.key)))
        headers["Authorization"] = "Basic %s" % (auth_b64.decode("utf-8"))

        action = API_PATH + "b2_authorize_account"
        resp = self.request(action=action, headers=headers, method="GET")

        if resp.status == httplib.OK:
            self._parse_and_set_auth_info(data=resp.object)
        else:
            raise Exception("Failed to authenticate: %s" % (str(resp.object)))

        return self

    def _parse_and_set_auth_info(self, data):
        result = {}
        self.account_id = data["accountId"]
        self.api_url = data["apiUrl"]
        self.download_url = data["downloadUrl"]
        self.auth_token = data["authorizationToken"]

        parsed_api_url = urlparse.urlparse(self.api_url)
        self.api_host = parsed_api_url.netloc

        parsed_download_url = urlparse.urlparse(self.download_url)
        self.download_host = parsed_download_url.netloc

        return result

    def _is_authentication_needed(self, force=False):
        if not self.auth_token or force:
            return True

        return False


class BackblazeB2Connection(ConnectionUserAndKey):
    host = None  # type: ignore  # Note: host is set after authentication
    secure = True
    responseCls = BackblazeB2Response
    authCls = BackblazeB2AuthConnection

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Stores info retrieved after authentication (auth token, api url,
        # download url).
        self._auth_conn = self.authCls(*args, **kwargs)

    def download_request(self, action, params=None):
        # Lazily perform authentication
        auth_conn = self._auth_conn.authenticate()

        # Set host to the download server
        self._set_host(auth_conn.download_host)

        action = "/file/" + action
        method = "GET"
        raw = True
        response = self._request(
            auth_conn=auth_conn, action=action, params=params, method=method, raw=raw
        )
        return response

    def upload_request(self, action, headers, upload_host, auth_token, data):
        # Lazily perform authentication
        auth_conn = self._auth_conn.authenticate()

        # Upload host is dynamically retrieved for each upload request
        self._set_host(host=upload_host)

        method = "POST"
        raw = False
        response = self._request(
            auth_conn=auth_conn,
            action=action,
            params=None,
            data=data,
            headers=headers,
            method=method,
            raw=raw,
            auth_token=auth_token,
        )
        return response

    def request(
        self,
        action,
        params=None,
        data=None,
        headers=None,
        method="GET",
        raw=False,
        include_account_id=False,
    ):
        params = params or {}
        headers = headers or {}

        # Lazily perform authentication
        auth_conn = self._auth_conn.authenticate()

        # Set host
        self._set_host(host=auth_conn.api_host)

        # Include Content-Type
        if not raw and data:
            headers["Content-Type"] = "application/json"

        # Include account id
        if include_account_id:
            if method == "GET":
                params["accountId"] = auth_conn.account_id
            elif method == "POST":
                data = data or {}
                data["accountId"] = auth_conn.account_id

        action = API_PATH + action
        if data:
            data = json.dumps(data)

        response = self._request(
            auth_conn=self._auth_conn,
            action=action,
            params=params,
            data=data,
            method=method,
            headers=headers,
            raw=raw,
        )
        return response

    def _request(
        self,
        auth_conn,
        action,
        params=None,
        data=None,
        headers=None,
        method="GET",
        raw=False,
        auth_token=None,
    ):
        params = params or {}
        headers = headers or {}

        if not auth_token:
            # If auth token is not explicitly provided, use the default one
            auth_token = self._auth_conn.auth_token

        # Include auth token
        headers["Authorization"] = "%s" % (auth_token)
        response = super().request(
            action=action,
            params=params,
            data=data,
            method=method,
            headers=headers,
            raw=raw,
        )
        return response

    def _set_host(self, host):
        """
        Dynamically set host which will be used for the following HTTP
        requests.

        NOTE: This is needed because Backblaze uses different hosts for API,
        download and upload requests.
        """
        self.host = host
        self.connection.host = "https://%s" % (host)


class BackblazeB2StorageDriver(StorageDriver):
    connectionCls = BackblazeB2Connection
    name = "Backblaze B2"
    website = "https://www.backblaze.com/b2/"
    type = Provider.BACKBLAZE_B2
    hash_type = "sha1"
    supports_chunked_encoding = False

    def iterate_containers(self):
        # pylint: disable=unexpected-keyword-arg
        resp = self.connection.request(
            action="b2_list_buckets", method="GET", include_account_id=True
        )
        containers = self._to_containers(data=resp.object)
        return containers

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :param prefix: Filter objects starting with a prefix.
                       Filtering is performed client-side.
        :type  prefix: ``str``

        :param ex_prefix: (Deprecated.) Filter objects starting with a prefix.
                          Filtering is performed client-side.
        :type  ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        # TODO: Support pagination
        params = {"bucketId": container.extra["id"]}
        resp = self.connection.request(action="b2_list_file_names", method="GET", params=params)
        objects = self._to_objects(data=resp.object, container=container)
        return self._filter_listed_container_objects(objects, prefix)

    def get_container(self, container_name):
        containers = self.iterate_containers()
        container = next((c for c in containers if c.name == container_name), None)
        if container:
            return container
        else:
            raise ContainerDoesNotExistError(value=None, driver=self, container_name=container_name)

    def get_object(self, container_name, object_name):
        container = self.get_container(container_name=container_name)
        objects = self.iterate_container_objects(container=container)

        obj = next((obj for obj in objects if obj.name == object_name), None)

        if obj is not None:
            return obj
        else:
            raise ObjectDoesNotExistError(value=None, driver=self, object_name=object_name)

    def create_container(self, container_name, ex_type="allPrivate"):
        data = {}
        data["bucketName"] = container_name
        data["bucketType"] = ex_type
        # pylint: disable=unexpected-keyword-arg
        resp = self.connection.request(
            action="b2_create_bucket", data=data, method="POST", include_account_id=True
        )
        container = self._to_container(item=resp.object)
        return container

    def delete_container(self, container):
        data = {}
        data["bucketId"] = container.extra["id"]
        # pylint: disable=unexpected-keyword-arg
        resp = self.connection.request(
            action="b2_delete_bucket", data=data, method="POST", include_account_id=True
        )
        return resp.status == httplib.OK

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        action = self._get_object_download_path(container=obj.container, obj=obj)
        # pylint: disable=no-member
        response = self.connection.download_request(action=action)

        # TODO: Include metadata from response headers
        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
            },
            success_status_code=httplib.OK,
        )

    def download_object_as_stream(self, obj, chunk_size=None):
        action = self._get_object_download_path(container=obj.container, obj=obj)
        # pylint: disable=no-member
        response = self.connection.download_request(action=action)

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={
                "iterator": response.iter_content(chunk_size),
                "chunk_size": chunk_size,
            },
            success_status_code=httplib.OK,
        )

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
    ):
        """
        Upload an object.

        Note: This will override file with a same name if it already exists.
        """
        # Note: We don't use any of the base driver functions since Backblaze
        # API requires you to provide SHA1 has upfront and the base methods
        # don't support that

        with open(file_path, "rb") as fp:
            iterator = iter(fp)
            iterator = read_in_chunks(iterator=iterator)
            data = exhaust_iterator(iterator=iterator)

        obj = self._perform_upload(
            data=data,
            container=container,
            object_name=object_name,
            extra=extra,
            verify_hash=verify_hash,
            headers=headers,
        )

        return obj

    def upload_object_via_stream(self, iterator, container, object_name, extra=None, headers=None):
        """
        Upload an object.

        Note: Backblaze does not yet support uploading via stream,
        so this calls upload_object internally requiring the object data
        to be loaded into memory at once
        """

        iterator = read_in_chunks(iterator=iterator)
        data = exhaust_iterator(iterator=iterator)

        obj = self._perform_upload(
            data=data,
            container=container,
            object_name=object_name,
            extra=extra,
            headers=headers,
        )

        return obj

    def delete_object(self, obj):
        data = {}
        data["fileName"] = obj.name
        data["fileId"] = obj.extra["fileId"]
        resp = self.connection.request(action="b2_delete_file_version", data=data, method="POST")
        return resp.status == httplib.OK

    def ex_get_object(self, object_id):
        params = {}
        params["fileId"] = object_id
        resp = self.connection.request(action="b2_get_file_info", method="GET", params=params)
        obj = self._to_object(item=resp.object, container=None)
        return obj

    def ex_hide_object(self, container_id, object_name):
        data = {}
        data["bucketId"] = container_id
        data["fileName"] = object_name
        resp = self.connection.request(action="b2_hide_file", data=data, method="POST")
        obj = self._to_object(item=resp.object, container=None)
        return obj

    def ex_list_object_versions(
        self,
        container_id,
        ex_start_file_name=None,
        ex_start_file_id=None,
        ex_max_file_count=None,
    ):
        params = {}
        params["bucketId"] = container_id

        if ex_start_file_name:
            params["startFileName"] = ex_start_file_name

        if ex_start_file_id:
            params["startFileId"] = ex_start_file_id

        if ex_max_file_count:
            params["maxFileCount"] = ex_max_file_count

        resp = self.connection.request(action="b2_list_file_versions", params=params, method="GET")
        objects = self._to_objects(data=resp.object, container=None)
        return objects

    def ex_get_upload_data(self, container_id):
        """
        Retrieve information used for uploading files (upload url, auth token,
        etc).

        :rype: ``dict``
        """
        # TODO: This is static (AFAIK) so it could be cached
        params = {}
        params["bucketId"] = container_id
        response = self.connection.request(action="b2_get_upload_url", method="GET", params=params)
        return response.object

    def ex_get_upload_url(self, container_id):
        """
        Retrieve URL used for file uploads.

        :rtype: ``str``
        """
        result = self.ex_get_upload_data(container_id=container_id)
        upload_url = result["uploadUrl"]
        return upload_url

    def _to_containers(self, data):
        result = []
        for item in data["buckets"]:
            container = self._to_container(item=item)
            result.append(container)

        return result

    def _to_container(self, item):
        extra = {}
        extra["id"] = item["bucketId"]
        extra["bucketType"] = item["bucketType"]
        container = Container(name=item["bucketName"], extra=extra, driver=self)
        return container

    def _to_objects(self, data, container):
        result = []
        for item in data["files"]:
            obj = self._to_object(item=item, container=container)
            result.append(obj)

        return result

    def _to_object(self, item, container=None):
        extra = {}
        extra["fileId"] = item["fileId"]
        extra["uploadTimestamp"] = item.get("uploadTimestamp", None)
        size = item.get("size", item.get("contentLength", None))
        hash = item.get("contentSha1", None)
        meta_data = item.get("fileInfo", {})
        obj = Object(
            name=item["fileName"],
            size=size,
            hash=hash,
            extra=extra,
            meta_data=meta_data,
            container=container,
            driver=self,
        )
        return obj

    def _get_object_download_path(self, container, obj):
        """
        Return a path used in the download requests.

        :rtype: ``str``
        """
        path = container.name + "/" + obj.name
        return path

    def _perform_upload(
        self, data, container, object_name, extra=None, verify_hash=True, headers=None
    ):
        if isinstance(data, str):
            data = bytearray(data)

        object_name = sanitize_object_name(object_name)

        extra = extra or {}
        content_type = extra.get("content_type", "b2/x-auto")
        meta_data = extra.get("meta_data", {})

        # Note: Backblaze API doesn't support chunked encoding and we need to
        # provide Content-Length up front (this is one inside _upload_object):/
        headers = headers or {}
        headers["X-Bz-File-Name"] = object_name
        headers["Content-Type"] = content_type

        sha1 = hashlib.sha1()
        sha1.update(b(data))
        headers["X-Bz-Content-Sha1"] = sha1.hexdigest()

        # Include optional meta-data (up to 10 items)
        for key, value in meta_data.items():
            # TODO: Encode / escape key
            headers["X-Bz-Info-%s" % (key)] = value

        upload_data = self.ex_get_upload_data(container_id=container.extra["id"])
        upload_token = upload_data["authorizationToken"]
        parsed_url = urlparse.urlparse(upload_data["uploadUrl"])

        upload_host = parsed_url.netloc
        request_path = parsed_url.path

        # pylint: disable=no-member
        response = self.connection.upload_request(
            action=request_path,
            headers=headers,
            upload_host=upload_host,
            auth_token=upload_token,
            data=data,
        )

        if response.status == httplib.OK:
            obj = self._to_object(item=response.object, container=container)
            return obj
        else:
            body = response.response.read()
            raise LibcloudError(
                "Upload failed. status_code={}, body={}".format(response.status, body),
                driver=self,
            )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import unittest

from libcloud.test.secrets import STORAGE_S3_PARAMS
from libcloud.storage.drivers.s3 import S3SignatureV4Connection
from libcloud.test.storage.test_s3 import S3Tests, S3MockHttp
from libcloud.storage.drivers.scaleway import SCW_FR_PAR_STANDARD_HOST, ScalewayStorageDriver


class ScalewayStorageDriverTestCase(S3Tests, unittest.TestCase):
    driver_type = ScalewayStorageDriver
    driver_args = STORAGE_S3_PARAMS
    default_host = SCW_FR_PAR_STANDARD_HOST

    @classmethod
    def create_driver(self):
        return self.driver_type(*self.driver_args, host=self.default_host)

    def setUp(self):
        super().setUp()

        ScalewayStorageDriver.connectionCls.conn_class = S3MockHttp
        S3MockHttp.type = None

        self.driver = self.create_driver()

    def test_connection_class_type(self):
        self.assertEqual(self.driver.connectionCls, S3SignatureV4Connection)

    def test_connection_class_default_host(self):
        self.assertEqual(self.driver.connectionCls.host, self.default_host)
        self.assertEqual(self.driver.connectionCls.port, 443)
        self.assertEqual(self.driver.connectionCls.secure, True)


if __name__ == "__main__":
    sys.exit(unittest.main())
